"""
 * Created by Synerty Pty Ltd
 *
 * This software is open source, the MIT license applies.
 *
 * Website : http://www.synerty.com
 * Support : support@synerty.com
"""

from twisted.trial import unittest

from vortex.Payload import Payload
from vortex.PayloadEndpoint import PayloadEndpoint
from vortex.PayloadEnvelope import PayloadEnvelope
from vortex.PayloadIO import PayloadIO


class PayloadEndpointPyTestAssignPayload():
    def __init__(self, unitTest):
        self._ut = unitTest

    def process(self, payloadEnvelope: PayloadEnvelope, **kwargs):
        self._ut.deliveredPayload = payloadEnvelope


class PayloadEndpointPyTest(unittest.TestCase):
    def setUp(self):
        PayloadIO._instance = None

        self.deliveredPayloadEnvelope = None

    def _testBuild(self, plFilt, epFilt):
        payload = Payload()
        payload.filt.update(plFilt)
        for x in range(6):
            payload.filt['%s' % x] = x

        def processPayload(payloadEnvelope: PayloadEnvelope, **kwargs):
            self.deliveredPayloadEnvelope = payloadEnvelope

        self._keepFuncInScope = processPayload

        PayloadEndpoint(epFilt, processPayload)

        PayloadIO().process(payloadEnvelope=payload.makePayloadEnvelope(),
                            vortexUuid='test',
                            vortexName='test',
                            httpSession=None,
                            sendResponse=lambda x: x)

        return payload

    def testFiltMatches(self):
        plFilt = {'key': 'unittest', 'This matches': 555}
        epFilt = {'key': 'unittest', 'This matches': 555}

        payload = self._testBuild(plFilt, epFilt)

        self.assertEqual(payload, self.deliveredPayloadEnvelope,
                         'PayloadIO/PayloadEndpoint delivery error')

    def testFiltValueUnmatched(self):
        plFilt = {'key': 'unittest', 'This matches': 555}
        epFilt = {'key': 'unittest', 'This matches': 0}

        self._testBuild(plFilt, epFilt)

        self.assertEqual(self.deliveredPayloadEnvelope, None,
                         'PayloadIO/PayloadEndpoint unmatched value test error')

    def testFiltKeyUnmatched(self):
        plFilt = {'key': 'unittest', 'This matches': 555}
        epFilt = {'key': 'unittest', 'This doesnt matches': 555}

        self._testBuild(plFilt, epFilt)

        self.assertEqual(self.deliveredPayloadEnvelope, None,
                         'PayloadIO/PayloadEndpoint unmatched value test error')

    def testFunctionGoesOutOfScope(self):
        filt = {'key': 'unittest', 'This matches': 555}

        payload = Payload()
        payload.filt = filt

        def subScope():
            def outOfScopeFunc(payloadEnvelope: PayloadEnvelope, *args, **kwargs):
                self.deliveredPayloadEnvelope = payloadEnvelope

            PayloadEndpoint(filt, outOfScopeFunc)

        PayloadIO().process(payloadEnvelope=payload.makePayloadEnvelope(),
                            vortexUuid='test',
                            vortexName='test',
                            httpSession=None,
                            sendResponse=lambda x: x)

        self.assertEqual(self.deliveredPayloadEnvelope, None,
                         'PayloadIO/PayloadEndpoint unmatched value test error')

    def testClassGoesOutOdScope(self):
        filt = {'key': 'unittest', 'This matches': 555}

        payload = Payload()
        payload.filt = filt

        def subScope():
            inst = PayloadEndpointPyTestAssignPayload(self)
            PayloadEndpoint(filt, inst.process)

        PayloadIO().process(payloadEnvelope=payload.makePayloadEnvelope(),
                            vortexUuid='test',
                            vortexName='test',
                            httpSession=None,
                            sendResponse=lambda x: x)

        self.assertEqual(self.deliveredPayloadEnvelope, None,
                         'PayloadIO/PayloadEndpoint unmatched value test error')

    def testClassStaysInScope(self):
        filt = {'key': 'unittest', 'This matches': 555}

        payload = Payload()
        payload.filt = filt

        inst = PayloadEndpointPyTestAssignPayload(self)
        PayloadEndpoint(filt, inst.process)

        PayloadIO().process(payloadEnvelope=payload.makePayloadEnvelope(),
                            vortexUuid='test',
                            vortexName='test',
                            httpSession=None,
                            sendResponse=lambda x: x)

        self.assertEqual(self.deliveredPayloadEnvelope, payload,
                         'PayloadIO/PayloadEndpoint unmatched value test error')



# import logging
#
# from vortex.PayloadFilterKeys import plDeleteKey
# from vortex.test.TestTuple import TestTuple
# from vortex.test.TupleDataForTest import makeTestTupleData
#
# logger = logging.getLogger(__name__)
#
#
# # If there were reallying going to a DB, we'd use OrmCrudHandler
# class VortexJSTupleLoaderTestHandler(ModelHandler):
#     def buildModel(self, payload, **kwargs):
#         logger.debug("Received payload with %s tuples and filt=%s",
#                      len(payload.tuples), payload.filt)
#
#         data = []
#
#         if payload.tuples:
#             # Return nothing if this was a delete
#             if plDeleteKey in payload.filt:
#                 # Return nothing, it was deleted
#                 pass
#
#             else:
#                 # Else this was a save, just update some data and return id
#                 data = payload.tuples
#                 for testTuple in data:
#                     testTuple.aInt = (testTuple.aInt if testTuple.aInt else 0) + 10
#                     testTuple.aBoolTrue = not (testTuple.aBoolTrue
#                                                if testTuple.aBoolTrue
#                                                else 0)
#
#         else:
#             # Else this is to get new data.
#             data = makeTestTupleData()
#
#         return data
#
#
# __handler = VortexJSTupleLoaderTestHandler({
#     "key": "vortex.tuple-loader.test.data"
# })



"""
 * Created by Synerty Pty Ltd
 *
 * This software is open source, the MIT license applies.
 *
 * Website : http://www.synerty.com
 * Support : support@synerty.com
"""
import uuid
from datetime import datetime
from typing import Dict
from typing import List

import pytz

from vortex.Tuple import Tuple
from vortex.Tuple import TupleField
from vortex.Tuple import addTupleType
from vortex.TupleSelector import TupleSelector


class TupleActionABC(Tuple):
    uuid: str = TupleField(
        comment="Uniquely generated id for this action", typingType=str
    )
    dateTime: datetime = TupleField(
        comment="The datetime this action was created", typingType=datetime
    )

    def __init__(self, **kwargs):
        if not self.__tupleType__:
            raise NotImplementedError(
                "TupleActionABC can not be instantiated,"
                " please inherit it and implement a tuple"
            )

        Tuple.__init__(self, **kwargs)

        if not "uuid" in kwargs:
            self.uuid = str(uuid.uuid1())  # Based on the host, and the time

        if not "dateTime" in kwargs:
            self.dateTime = datetime.now(pytz.utc)


@addTupleType
class TupleGenericAction(TupleActionABC):
    """Tuple Generic Action

    This is a generic action, to be used when the implementor doesn't want to implement
    concrete classes for each action type.

    """

    __tupleType__ = "vortex.TupleGenericAction"

    key = TupleField(comment="An optional key for this action", typingType=str)

    data = TupleField(comment="Optional data for the update", typingType=str)


@addTupleType
class TupleUpdateAction(TupleActionABC):
    __tupleType__ = "vortex.TupleUpdateAction"

    tupleSelector = TupleField(
        comment="The tuple selector for this action", typingType=TupleSelector
    )
    tupleChanges = TupleField(
        comment="An array of {old:v,new:v} dicts for the changes", typingType=List[Dict]
    )



"""
 * Created by Synerty Pty Ltd
 *
 * This software is open source, the MIT license applies.
 *
 * Website : http://www.synerty.com
 * Support : support@synerty.com
"""
import logging
import ujson
import zlib
from base64 import b64encode, b64decode
from datetime import datetime
from typing import List, Optional, Dict

import pytz

from .DeferUtil import deferToThreadWrapWithLogger, noMainThread
from .Payload import Payload
from .SerialiseUtil import T_RAPUI_PAYLOAD_ENVELOPE
from .Jsonable import Jsonable

logger = logging.getLogger(__name__)

PayloadEnvelopeList = List['PayloadEnvelope']
VortexMsgList = List[bytes]


class NoPayloadException(Exception):
    pass


class PayloadEnvelope(Jsonable):
    ''' Payload
    This object represents a hierarchy of data transferred between client and server
    '''
    __fieldNames__ = ['filt', 'encodedPayload', 'result', 'date']
    __rapuiSerialiseType__ = T_RAPUI_PAYLOAD_ENVELOPE

    vortexUuidKey = '__vortexUuid__'
    vortexNameKey = '__vortexName__'

    def __init__(self, filt: Optional[Dict] = None,
                 encodedPayload: Optional[bytes] = None,
                 result=None,
                 date: Optional[datetime] = None) -> None:
        """ Constructor

        """
        self.filt: Dict = {} if filt is None else filt
        self.encodedPayload: bytes = encodedPayload
        self.result = result
        self.date: datetime = date if date else datetime.now(pytz.utc)

        if isinstance(filt, str):
            self.filt = {self.filt: None}

    # -------------------------------------------
    # PayloadEnvelope Methods

    def isEmpty(self):
        return ((not self.filt
                 or (self.vortexNameKey in self.filt
                     and self.vortexUuidKey in self.filt
                     and len(self.filt) == 2))
                and not self.encodedPayload
                and not self.result)

    def decodePayload(self) -> Payload:
        if not self.encodedPayload:
            raise NoPayloadException()

        noMainThread()
        return Payload().fromEncodedPayload(self.encodedPayload)

    @deferToThreadWrapWithLogger(logger)
    def decodePayloadDefer(self) -> "Payload":
        return self.decodePayload()

    # -------------------------------------------
    # JSON Related methods
    def _fromJson(self, jsonStr):
        jsonDict = ujson.loads(jsonStr)

        assert (jsonDict[Jsonable.JSON_CLASS_TYPE] == self.__rapuiSerialiseType__)
        return self.fromJsonDict(jsonDict)

    def _toJson(self) -> str:
        return ujson.dumps(self.toJsonDict())

    # -------------------------------------------
    # VortexServer Message Methods
    def toVortexMsg(self) -> bytes:
        return b64encode(self._toJson().encode())

    @deferToThreadWrapWithLogger(logger)
    def toVortexMsgDefer(self) -> bytes:
        return self.toVortexMsg()

    def fromVortexMsg(self, vortexMsg: bytes) -> "PayloadEnvelope":
        jsonStr = b64decode(vortexMsg).decode()
        return self._fromJson(jsonStr)

    @deferToThreadWrapWithLogger(logger)
    def fromVortexMsgDefer(self, vortexMsg: bytes) -> "PayloadEnvelope":
        return self.fromVortexMsg(vortexMsg)



# import unittest
#
#
# class MyTestCase(unittest.TestCase):
#     def test_something(self):
#         self.assertEqual(True, False)
#
#
# if __name__ == '__main__':
#     unittest.main()
from vortex.rpc.RPC import vortexRPC


@vortexRPC("listenVortexName", acceptOnlyFromVortex="sendVortexName")
def myRemoteAddMethod(arg1: int, kwarg1: int = 0) -> int:
    raise ValueError("A is not B")



"""
 * Created by Synerty Pty Ltd
 *
 * This software is open source, the MIT license applies.
 *
 * Website : http://www.synerty.com
 * Support : support@synerty.com
"""
import logging
import uuid
from datetime import datetime
from typing import Union, Optional, List

import pytz
import twisted
from twisted.internet import reactor
from twisted.internet import task
from twisted.internet.defer import Deferred, inlineCallbacks
from twisted.internet.error import ConnectionDone, ConnectionRefusedError
from twisted.internet.protocol import connectionDone, ReconnectingClientFactory

from vortex.DeferUtil import vortexLogFailure, isMainThread
from vortex.PayloadEnvelope import PayloadEnvelope, VortexMsgList
from vortex.PayloadPriority import DEFAULT_PRIORITY
from vortex.VortexABC import VortexABC, VortexInfo
from vortex.VortexPayloadProtocol import VortexPayloadProtocol
from vortex.VortexServer import HEART_BEAT_PERIOD
from vortex.VortexWritePushProducer import VortexWritePushProducer

logger = logging.getLogger(name=__name__)


class VortexPayloadTcpClientProtocol(VortexPayloadProtocol):
    def __init__(self, vortexClient=None) -> None:
        VortexPayloadProtocol.__init__(self, logger)
        self._vortexClient = vortexClient

        self._closed = False

        # Start our heart beat
        self._sendBeatLoopingCall = task.LoopingCall(self._sendBeat)
        d = self._sendBeatLoopingCall.start(HEART_BEAT_PERIOD, now=False)
        d.addErrback(lambda f: logger.exception(f.value))

        self._producer: Optional[VortexWritePushProducer] = None

    def _beat(self):
        if self._vortexClient:
            self._vortexClient._beat()

    def _nameAndUuidReceived(self, name, uuid):
        from vortex.VortexFactory import VortexFactory
        VortexFactory._notifyOfVortexStatusChange(name, online=True)

        self._producer.setRemoteVortexName(self._serverVortexName)

        if self._vortexClient:
            self._vortexClient._setNameAndUuid(name=self._serverVortexName,
                                               uuid=self._serverVortexUuid)

    def _createResponseSenderCallable(self):
        def sendResponse(vortexMsgs: Union[VortexMsgList, bytes],
                         priority: int = DEFAULT_PRIORITY):
            return self._vortexClient.sendVortexMsg(vortexMsgs=vortexMsgs,
                                                    priority=priority)

        return sendResponse

    def _sendBeat(self):
        if self._closed:
            return

        # Send the heartbeats
        self._producer.write(b'.', DEFAULT_PRIORITY)

    def write(self, payloadVortexStr: bytes, priority: int = DEFAULT_PRIORITY):
        if not twisted.python.threadable.isInIOThread():
            e = Exception("Write called from NON main thread")
            logger.exception(str(e))
            raise e

        assert not self._closed
        self._producer.write(payloadVortexStr, priority)

    def connectionMade(self):
        self._producer = VortexWritePushProducer(self.transport, lambda: self.close())

        # Register the producer if there isn't one already.
        if not self.transport.producer:
            self.transport.registerProducer(self._producer, True)

        # Send a heart beat down the new connection, tell it who we are.
        connectPayloadFilt = {
            PayloadEnvelope.vortexUuidKey: self._vortexClient.uuid,
            PayloadEnvelope.vortexNameKey: self._vortexClient.name
        }
        self._producer.write(PayloadEnvelope(filt=connectPayloadFilt).toVortexMsg(),
                             DEFAULT_PRIORITY)

    def connectionLost(self, reason=connectionDone):
        from vortex.VortexFactory import VortexFactory
        VortexFactory._notifyOfVortexStatusChange(self._serverVortexName, online=False)

        if self._sendBeatLoopingCall.running:
            self._sendBeatLoopingCall.stop()
        self._closed = False

    def close(self):
        self.transport.loseConnection()
        if self._sendBeatLoopingCall.running:
            self._sendBeatLoopingCall.stop()
        self._closed = False


class VortexClientTcp(ReconnectingClientFactory, VortexABC):
    """ VortexServer Client
    Connects to a votex server
    """
    RETRY_DELAY = 1.5  # Seconds
    HEART_BEAT_TIMEOUT = 30.0  # Seconds

    # The time it takes after recieving a response from the server to receive the
    INFO_PAYLOAD_TIMEOUT = 5  # Seconds

    # Set the ReconnectingClientFactory max delay
    maxDelay = 1

    def __init__(self, name: str) -> None:
        self._vortexName = name
        self._vortexUuid = str(uuid.uuid1())

        self._server = None
        self._port = None

        self._retrying = False

        self._serverVortexUuid = None
        self._serverVortexName = None

        self._lastBeatReceiveTime = None
        self._lastHeartBeatCheckTime = datetime.now(pytz.utc)

        # Start our heart beat checker
        self._beatLoopingCall = task.LoopingCall(self._checkBeat)

        self._reconnectVortexMsgs = [PayloadEnvelope().toVortexMsg()]

        self.__protocol = None

    #####################################################################################3

    # class EchoClientFactory(ReconnectingClientFactory):
    def startedConnecting(self, connector):
        logger.debug('Started to connect.')

    def buildProtocol(self, addr):

        if self.__protocol:
            self.__protocol.close()
            self.__protocol = None

            logger.debug('Connected.')
        self.resetDelay()
        self.__protocol = VortexPayloadTcpClientProtocol(self)
        return self.__protocol

    def clientConnectionLost(self, connector, reason):
        if not reason.check(ConnectionDone):
            logger.debug('Lost connection.  Reason: %s', reason)
        ReconnectingClientFactory.clientConnectionLost(self, connector, reason)

    def clientConnectionFailed(self, connector, reason):
        if reason.check(ConnectionRefusedError):
            logger.debug("Connection refused (We'll retry)")
        else:
            logger.debug('Connection failed. Reason: %s', reason)
        ReconnectingClientFactory.clientConnectionFailed(self, connector,
                                                         reason)

    #####################################################################################3

    @property
    def localVortexInfo(self) -> VortexInfo:
        return VortexInfo(name=self._vortexName,
                          uuid=self._vortexUuid)

    @property
    def remoteVortexInfo(self) -> List[VortexInfo]:
        if not self.__protocol:
            return []

        if not self._serverVortexUuid:
            return []

        return [VortexInfo(name=self._serverVortexName,
                           uuid=self._serverVortexUuid)]

    @property
    def name(self):
        return self._vortexName

    @property
    def uuid(self):
        return self._vortexUuid

    def connect(self, server, port):
        self._server = server
        self._port = port

        self._beat()
        d = self._beatLoopingCall.start(5.0, now=False)
        d.addErrback(vortexLogFailure, logger, consumeError=True)

        deferred = Deferred()

        reactor.connectTCP(self._server, self._port, self)

        def checkUuid():
            if self._serverVortexName:
                deferred.callback(True)
            else:
                reactor.callLater(0.1, checkUuid)

        checkUuid()

        return deferred

    def close(self):
        if self._beatLoopingCall and self._beatLoopingCall.running:
            self._beatLoopingCall.stop()
            self._beatLoopingCall = None

        self.stopTrying()
        self.__protocol.close()

    def addReconnectVortexMsg(self, vortexMsg: bytes):
        """ Add Reconnect Payload
        :param vortexMsg: An encoded PayloadEnvelope to send when the connection reconnects
        :return:
        """
        self._reconnectVortexMsgs.append(vortexMsg)

    def sendVortexMsg(self,
                      vortexMsgs: Union[VortexMsgList, bytes, None] = None,
                      vortexUuid: Optional[str] = None,
                      priority: int = DEFAULT_PRIORITY) -> Deferred:
        """ Send Vortex Msg

        NOTE: Priority ins't supported as there is no buffer for this class.

        """

        if vortexMsgs is None:
            vortexMsgs = self._reconnectVortexMsgs

        if not isinstance(vortexMsgs, list):
            vortexMsgs = [vortexMsgs]

        # Check if the vortexUuid matches the destination uuid
        #
        # if vortexUuid and vortexUuid != self._
        #
        # if not self.__protocol.serverVortexUuid:
        #     return []

        if isMainThread():
            return self._sendVortexMsgLater(vortexMsgs)

        return task.deferLater(reactor, 0, self._sendVortexMsgLater, vortexMsgs)

    @inlineCallbacks
    def _sendVortexMsgLater(self, vortexMsgs: VortexMsgList):
        yield None
        assert self._server
        assert vortexMsgs

        self.vortexMsgs = b''

        for vortexMsg in vortexMsgs:
            self.__protocol.write(vortexMsg)

        return True

    def _beat(self):
        """ Beat, Called by protocol """
        self._lastBeatReceiveTime = datetime.now(pytz.utc)

    def _setNameAndUuid(self, name, uuid):
        """ Set Name And Uuid, Called by protocol """
        self._serverVortexName = name
        self._serverVortexUuid = uuid

    def _checkBeat(self):

        # If we've been asleep, then make note of that (VM suspended)
        checkTimout = (datetime.now(pytz.utc) - self._lastHeartBeatCheckTime) \
                          .seconds > self.HEART_BEAT_TIMEOUT

        # Has the heart beat expired?
        beatTimeout = (datetime.now(pytz.utc) - self._lastBeatReceiveTime) \
                          .seconds > self.HEART_BEAT_TIMEOUT

        # Mark that we've just checked it
        self._lastHeartBeatCheckTime = datetime.now(pytz.utc)

        if checkTimout:
            self._lastBeatReceiveTime = datetime.now(pytz.utc)
            return

        if beatTimeout:
            self._reconnectAfterHeartBeatLost()
            return

    def _reconnectAfterHeartBeatLost(self):
        if self._retrying:
            return

        self._retrying = True

        logger.info("VortexServer client dead, reconnecting %s:%s",
                    self._server, self._port)

        if self.__protocol:
            self.__protocol.close()



import json
import logging
from enum import Enum
from typing import Callable
from typing import List

from twisted.internet.defer import inlineCallbacks
from twisted.python.compat import nativeString
from twisted.web.server import NOT_DONE_YET
from txhttputil.site.BasicResource import BasicResource
from vortex import Tuple
from vortex.DeferUtil import deferToThreadWrapWithLogger
from vortex.DeferUtil import vortexLogFailure

logger = logging.getLogger(__name__)


class HTTP_REQUEST(Enum):
    GET = "GET"
    POST = "POST"
    PUT = "PUT"
    DELETE = "DELETE"


class _JsonResource(BasicResource):
    isLeaf = True
    isGzipped = True

    def __init__(self, handler: Callable, TupleClass: Tuple):
        BasicResource.__init__(self)

        self._handler = handler
        self._tupleClass = TupleClass

    @deferToThreadWrapWithLogger(logger)
    def _requestToTuple(self, request) -> Tuple:
        req = json.load(request.content)
        tuple_ = self._tupleClass()
        tuple_.fromJsonDict(req)
        return tuple_

    @deferToThreadWrapWithLogger(logger)
    def _tupleToResponse(self, tuple_: Tuple) -> bytes:
        dict_ = tuple_.tupleToRestfulJsonDict()
        json_ = json.dumps(dict_)
        return json_.encode()

    def _writeSuccessResponse(self, response: bytes, request):
        request.setResponseCode(200)
        request.setHeader("Content-Type", "application/json")
        request.write(response)
        request.finish()

    def _writeErrorResponse(self, request):
        request.setResponseCode(500)
        request.setHeader("Content-Type", "application/json")
        request.write(b'{"error":"internal error"}')
        request.finish()

    @inlineCallbacks
    def _renderAsync(self, request):
        tupleIn = None
        try:
            tupleIn = yield self._requestToTuple(request)
            tupleOut = yield self._handler(tupleIn)
            response = yield self._tupleToResponse(tupleOut)
            self._writeSuccessResponse(response, request)

        except Exception as e:
            self._writeErrorResponse(request)
            logger.debug(
                f"Error while processing REST Tuple\n{tupleIn}\n{request.path}"
            )
            logger.exception(e)

    def render(self, request):
        d = self._renderAsync(request)
        d.addErrback(vortexLogFailure, logger=logger, consumeError=True)

        return NOT_DONE_YET


class ErrorJsonResource(BasicResource):
    isLeaf = True
    isGzipped = True

    def __init__(self, errorCode: int):
        self._errorCode = errorCode

    def render(self, request):
        request.setResponseCode(self._errorCode)
        return b""


class PluginRestfulResource(BasicResource):
    def __init__(self):
        BasicResource.__init__(self)
        self._registeredMethods = {}

    def registerMethod(
        self,
        handlerFunction: Callable,
        TupleClass: Tuple,
        urlPrefix: bytes,
        registeredRequestMethod: List[HTTP_REQUEST],
    ):
        if urlPrefix in self._registeredMethods:
            raise ValueError(f'Route "{urlPrefix}" already exists')
        if not callable(handlerFunction):
            raise TypeError(f'"{str(handlerFunction)}" is not a callable')

        # register handler as partial function (poor man's Template<type>)
        # as callback function whose input is an instance of a Tuple
        self._registeredMethods[urlPrefix] = {
            "handler": handlerFunction,
            "allowedMethods": set([method.value for method in registeredRequestMethod]),
            "tupleClass": TupleClass,
        }

    def getChild(self, path, request):
        """Get requests routed to an handler

        Routes to a registered handler to process the request.
        This function routes resources to handlers which eventually returns
        JSON response resource in json. Everything returned here should be
        final.

        Once url prefix for a resource is matched, it checks the
        request method is allowed. If granted, it delegates the request to the
        registered handler to process the request. The handler should convert
        request to a Tuple and should respond a tuple.

        :param path: route url for a resource, with plugin name prefix stripped
        :param request: twisted.web.http.Request
        :return: twisted.web.resource.Resource
        """
        # route check
        if path not in self._registeredMethods.keys():
            return ErrorJsonResource(404)

        # request method check
        requestMethod = nativeString(request.method)
        if requestMethod not in self._registeredMethods[path]["allowedMethods"]:
            return ErrorJsonResource(403)

        # invoke resource handler
        TupleClass = self._registeredMethods[path]["tupleClass"]

        handler = self._registeredMethods[path]["handler"]

        return _JsonResource(handler, TupleClass)



"""
 * Created by Synerty Pty Ltd
 *
 * This software is open source, the MIT license applies.
 *
 * Website : http://www.synerty.com
 * Support : support@synerty.com
"""

from . import TupleActionProcessorTestHandler
from . import TupleDataObservableTestHandler
# from . import VortexJSTupleLoaderTestHandler



"""
 * Created by Synerty Pty Ltd
 *
 * This software is open source, the MIT license applies.
 *
 * Website : http://www.synerty.com
 * Support : support@synerty.com
"""

import datetime
import json
from typing import Dict, Annotated
from typing import List
from typing import Union

from ddt import data, ddt
from twisted.trial import unittest

from vortex.Tuple import TUPLE_TYPES_BY_NAME, TupleHash, IntTupleFieldValidator
from vortex.Tuple import Tuple
from vortex.Tuple import TupleField
from vortex.Tuple import addTupleType


@addTupleType
class RestfulTestTuple(Tuple):
    """
    classdocs
    """

    __tupleType__ = "synerty.vortex.RestfulTestTuple"
    aInt: int = TupleField()
    aIntWithValidator: Annotated[int, IntTupleFieldValidator(1, 100)] = TupleField()
    aFloat: float = TupleField()
    aString: str = TupleField()
    aBoolTrue: bool = TupleField()
    aBoolFalse: bool = TupleField()
    aList: List[Union[str, int, bool, float]] = TupleField()
    aDict: Dict[str, Union[str, float]] = TupleField()
    aSubTuple: "RestfulSubTestTuple" = TupleField()
    aListOfSubTuples: List["RestfulSubTestTuple"] = TupleField()
    aDictOfSubTuples: Dict[str, "RestfulSubTestTuple"] = TupleField()


@addTupleType
class RestfulSubTestTuple(Tuple):
    """Test Sub Tuple
    This tuple will test if we can have tuples in tuples and still serialise and
    deserialise them.
    """

    __tupleType__ = "rapui.synerty.RestfulSubTestTuple"
    subInt: int = TupleField()


def _makeRestfultTestTuple():
    tuple_ = RestfulTestTuple()
    tuple_.aBoolTrue = True
    tuple_.aBoolFalse = False
    tuple_.aFloat = 2.56
    tuple_.aInt = 1231231
    tuple_.aIntWithValidator = 50
    tuple_.aString = "test string from 345345345@$#%#$%#$%#$%#"
    tuple_.aList = ["y", 3, True, 3.3]
    tuple_.aDict = {
        "a": "test str",
        "1": 1.234,
    }  # Key strings are loaded as unicode
    tuple_.aDictOfSubTuples = {}

    subTuple = RestfulSubTestTuple()
    subTuple.subInt = 2048
    tuple_.aSubTuple = subTuple

    tuple_.aListOfSubTuples = []
    for x in range(2):
        subTuple = RestfulSubTestTuple()
        subTuple.subInt = x
        tuple_.aListOfSubTuples.append(subTuple)

        subTuple = RestfulSubTestTuple()
        subTuple.subInt = x + 10
        tuple_.aDictOfSubTuples[str(x)] = subTuple

    return tuple_


@ddt
class TupleRestfulTest(unittest.TestCase):
    def setUp(self):
        pass

    def tearDown(self):
        pass

    def _testSerialisation(self, origTuple):
        # To Restful JSON Object
        jsonDict = origTuple.tupleToRestfulJsonDict()
        # Pass it through the JSON object to JSON string conversion
        jsonDict = json.loads(json.dumps(jsonDict))
        # Reconstruct a new Tuple
        deserialisedTuple = RestfulTestTuple().restfulJsonDictToTupleWithValidation(
            jsonDict, RestfulTestTuple
        )
        return deserialisedTuple

    @data(-1, 110)
    def testIntValidatorInvalid(self, int_):
        # Create tuple
        origTuple = _makeRestfultTestTuple()
        origTuple.aIntWithValidator = int_
        self.assertRaises(ValueError, self._testSerialisation, origTuple)

    @data(1, 99)
    def testIntValidatorValid(self, int_):
        # Create tuple
        origTuple = _makeRestfultTestTuple()
        origTuple.aIntWithValidator = int_
        self.assertEqual(self._testSerialisation(origTuple).aIntWithValidator, int_)

    def testValidatorPass(self):
        # Create tuple
        origTuple = _makeRestfultTestTuple()

        deserialisedTuple = self._testSerialisation(origTuple)

        self.assertEqual(TupleHash(origTuple), TupleHash(origTuple))
        self.assertEqual(TupleHash(deserialisedTuple), TupleHash(deserialisedTuple))
        self.assertEqual(
            TupleHash(origTuple),
            TupleHash(deserialisedTuple),
            "Tuple serialise -> deserialise error",
        )



"""
 * Created by Synerty Pty Ltd
 *
 * This software is open source, the MIT license applies.
 *
 * Website : http://www.synerty.com
 * Support : support@synerty.com
"""
import logging
from copy import copy
from typing import Callable, List
from typing import Union

from sqlalchemy.orm.exc import NoResultFound
from sqlalchemy.orm.session import Session
from twisted.internet.defer import inlineCallbacks
from twisted.internet.threads import deferToThread

from vortex.DeferUtil import deferToThreadWrapWithLogger
from vortex.Payload import Payload
from vortex.PayloadEndpoint import PayloadEndpoint
from vortex.PayloadEnvelope import VortexMsgList, PayloadEnvelope
from vortex.PayloadFilterKeys import plIdKey, plDeleteKey
from vortex.Tuple import Tuple
from vortex.VortexFactory import VortexFactory

logger = logging.getLogger(__name__)


class OrmCrudHandlerExtension(object):
    def uiData(self, tuple_, tuples, session, payloadFilt):
        return True

    def afterCreate(self, tuple_, tuples, session, payloadFilt):
        return True

    def afterRetrieve(self, tuple_, tuples, session, payloadFilt):
        return True

    def beforeUpdate(self, tuple_, tuples, session, payloadFilt):
        return True

    def middleUpdate(self, tuple_, tuples, session, payloadFilt):
        return True

    def afterUpdate(self, tuple_, tuples, session, payloadFilt):
        return True

    def afterUpdateCommit(self, tuple_, tuples, session, payloadFilt):
        return True

    def beforeDelete(self, tuple_, tuples, session, payloadFilt):
        return True

    def afterDeleteCommit(self, tuple_, tuples, session, payloadFilt):
        return True


class _OrmCrudExtensionProcessor(object):
    def __init__(self):
        self.extensions = {}

    def addExtensionClassDecorator(self, Tuple):
        def f(cls):
            self.extensions[Tuple.tupleType()] = cls()
            return cls

        return f

    def addExtensionObject(self, Tuple, ormCrudHandlerExtension):
        self.extensions[Tuple.tupleType()] = ormCrudHandlerExtension

    def uiData(self, tuples, session, payloadFilt):
        if not self.extensions:
            return

        for tuple_ in tuples:
            if not hasattr(tuple_, 'uiData'):
                continue
            extension = self.extensions.get(tuple_.tupleType())
            if extension:
                extension.uiData(tuple_, tuples, session, payloadFilt)

    def afterCreate(self, tuples, session, payloadFilt):
        if not self.extensions:
            return

        for tuple_ in tuples:
            extension = self.extensions.get(tuple_.tupleType())
            if extension:
                extension.afterCreate(tuple_, tuples, session, payloadFilt)

        self.uiData(tuples, session, payloadFilt)

    def afterRetrieve(self, tuples, session, payloadFilt):
        if not self.extensions:
            return

        for tuple_ in tuples:
            extension = self.extensions.get(tuple_.tupleType())
            if extension:
                extension.afterRetrieve(tuple_, tuples, session, payloadFilt)

        self.uiData(tuples, session, payloadFilt)

    def beforeUpdate(self, tuples, session, payloadFilt):
        if not self.extensions:
            return

        for tuple_ in tuples:
            extension = self.extensions.get(tuple_.tupleType())
            if extension:
                extension.beforeUpdate(tuple_, tuples, session, payloadFilt)

    def middleUpdate(self, tuples, session, payloadFilt):
        if not self.extensions:
            return

        for tuple_ in tuples:
            extension = self.extensions.get(tuple_.tupleType())
            if extension:
                extension.middleUpdate(tuple_, tuples, session, payloadFilt)

    def afterUpdate(self, tuples, session, payloadFilt):
        if not self.extensions:
            return

        for tuple_ in tuples:
            extension = self.extensions.get(tuple_.tupleType())
            if extension:
                extension.afterUpdate(tuple_, tuples, session, payloadFilt)

    def afterUpdateCommit(self, tuples, session, payloadFilt):
        if not self.extensions:
            return

        for tuple_ in tuples:
            extension = self.extensions.get(tuple_.tupleType())
            if extension:
                extension.afterUpdateCommit(tuple_, tuples, session, payloadFilt)

        self.uiData(tuples, session, payloadFilt)

    def beforeDelete(self, tuples, session, payloadFilt):
        if not self.extensions:
            return

        for tuple_ in tuples:
            extension = self.extensions.get(tuple_.tupleType())
            if extension:
                extension.beforeDelete(tuple_, tuples, session, payloadFilt)

    def afterDeleteCommit(self, tuples, session, payloadFilt):
        if not self.extensions:
            return

        for tuple_ in tuples:
            extension = self.extensions.get(tuple_.tupleType())
            if extension:
                extension.afterDeleteCommit(tuple_, tuples, session, payloadFilt)


class OrmCrudHandler(object):
    UPDATE = 1
    CREATE = 2
    DELETE = 3
    QUERY = 4

    def __init__(self, sessionFunctor, Declarative, payloadFilter, retreiveAll=False):
        ''' Create CRUD Hanlder

        This handler will perform crud operations for an SQLAlchemy Declarative

        '''
        self._sessionFunctor = sessionFunctor
        self._Declarative = Declarative
        self._payloadFilter = (payloadFilter
                               if isinstance(payloadFilter, dict) else
                               {"key": payloadFilter})

        self._ep = PayloadEndpoint(self._payloadFilter, self._process)

        self._ext = _OrmCrudExtensionProcessor()

        self._retreiveAll = retreiveAll

    def shutdown(self):
        self._ep.shutdown()

    def addExtension(self, Tuple, ormCrudHandlerExtension=None):
        if ormCrudHandlerExtension:
            self._ext.addExtensionObject(Tuple, ormCrudHandlerExtension)
            return
        return self._ext.addExtensionClassDecorator(Tuple)

    @inlineCallbacks
    def _process(self, *args, **kwargs):
        val = yield self._processInThread(*args, **kwargs)
        return val

    @deferToThreadWrapWithLogger(logger)
    def _processInThread(self, payloadEnvelope: PayloadEnvelope,
                         vortexUuid: str,
                         sendResponse: Callable[[Union[VortexMsgList, bytes]], None],
                         **kwargs):

        # Execute preprocess functions
        if self.preProcess(payloadEnvelope, vortexUuid, **kwargs) != None:
            return

        # Create reply payload replyFilt

        replyFilt = copy(self._payloadFilter)
        if payloadEnvelope.filt:
            replyFilt.update(payloadEnvelope.filt)

        # Get data from the payload
        phId = payloadEnvelope.filt.get(plIdKey)
        delete = payloadEnvelope.filt.get(plDeleteKey, False)

        # Setup variables to populate
        replyPayloadEnvelope: PayloadEnvelope = None
        action = None

        tuples = []
        if payloadEnvelope.encodedPayload:
            tuples = payloadEnvelope.decodePayload().tuples

        session = self._getSession()

        # Execute the action
        try:
            if delete == True:
                action = self.DELETE
                replyPayloadEnvelope = self._delete(session, tuples, phId,
                                                    payloadEnvelope.filt)

            elif len(tuples):
                action = self.UPDATE
                replyPayloadEnvelope = self._update(session, tuples,
                                                    payloadEnvelope.filt)

            elif phId != None:
                action = self.QUERY
                replyPayloadEnvelope = self._retrieve(session, phId, payloadEnvelope.filt)

            elif len(tuples) == 0:
                action = self.CREATE
                replyPayloadEnvelope = self._create(session, payloadEnvelope.filt)

            else:
                session.close()
                raise Exception("Invalid ORM CRUD parameter state")

        except Exception as e:
            replyPayloadEnvelope = PayloadEnvelope(result=str(e), filt=replyFilt)
            sendResponse(replyPayloadEnvelope.toVortexMsg())
            try:
                session.rollback()
            except:
                pass
            session.close()
            raise

        # Prefer reply filt, if not combine our accpt filt with the filt we were sent

        # Ensure any delegates are playing nice with the result
        if action in (self.DELETE, self.UPDATE) and replyPayloadEnvelope.result is None:
            replyPayloadEnvelope.result = True

        replyPayloadEnvelope.filt = replyFilt
        sendResponse(replyPayloadEnvelope.toVortexMsg())

        # Execute the post process function
        self.postProcess(action, payloadEnvelope.filt, vortexUuid)
        session.commit()
        session.close()

    def _getSession(self):
        if isinstance(self._sessionFunctor, Session):
            return self._sessionFunctor

        return self._sessionFunctor()

    def _getDeclarativeById(self, session, id_):
        qry = session.query(self._Declarative)
        if self._retreiveAll and id_ is None:
            return qry.all()

        try:
            return qry.filter(self._Declarative.id == id_).one()
        except NoResultFound as e:
            return None

    def createDeclarative(self, session, payloadFilt):
        if self._retreiveAll:
            return session.query(self._Declarative).all()

        return [self._Declarative()]

    def _getDeclarativeByTuple(self, session, tuple_):
        T = tuple_.__class__
        return session.query(T).filter(T.id == tuple_.id).one()

    def _create(self, session, payloadFilt) -> PayloadEnvelope:
        tuples = self.createDeclarative(session, payloadFilt)
        payload = Payload(tuples=tuples)
        self._ext.afterCreate(payload.tuples, session, payloadFilt)
        return payload.makePayloadEnvelope()

    def _retrieve(self, session, filtId, payloadFilt, obj=None,
                  **kwargs) -> PayloadEnvelope:
        ph = obj if obj else self._getDeclarativeById(session, filtId)
        payload = Payload()
        payload.tuples = [ph] if ph else []
        self._ext.afterRetrieve(payload.tuples, session, payloadFilt)
        return payload.makePayloadEnvelope()

    def _update(self, session, tuples, payloadFilt) -> PayloadEnvelope:
        self._ext.beforeUpdate(tuples, session, payloadFilt)

        # Add everything first.
        for tupleObj in tuples:
            if tupleObj is None:
                raise Exception("None/null was present in array of tuples")

            # Make sure it's not ''
            if tupleObj.id == '':
                tupleObj.id = None

            if tupleObj.id is None:
                session.add(tupleObj)

        self._ext.middleUpdate(tuples, session, payloadFilt)

        # Now merge with the session
        returnTuples = []
        for tupleObj in tuples:
            if tupleObj.id is None:
                # If this was a create, then we can just return this tuple
                returnTuples.append(tupleObj)

            else:
                # Otherwise use the merge method to perform the update magic.
                # and add the resulting merged tuple to the return list
                returnTuples.append(session.merge(tupleObj))

        self._ext.afterUpdate(returnTuples, session, payloadFilt)

        session.commit()

        self._ext.afterUpdateCommit(returnTuples, session, payloadFilt)

        return Payload(tuples=returnTuples).makePayloadEnvelope(result=True)

    def _delete(self, session, tuples, filtId, payloadFilt) -> PayloadEnvelope:
        self._ext.beforeDelete(tuples, session, payloadFilt)

        if len(tuples):
            phIds = [t.id for t in tuples]
        else:
            phIds = [filtId]

        for phId in phIds:
            ph = self._getDeclarativeById(session, phId)
            try:
                # Try to iterate it
                for item in iter(ph):
                    session.delete(item)

            except TypeError:
                # If it's not an iterator
                if ph is not None:
                    session.delete(ph)

        session.commit()

        returnTuples: List[Tuple] = []
        if self._retreiveAll:
            returnTuples = self.createDeclarative(session, payloadFilt)

        self._ext.afterDeleteCommit(tuples, session, payloadFilt)
        return Payload(tuples=returnTuples).makePayloadEnvelope(result=True)

    def sendModelUpdate(self, objId,
                        vortexUuid=None,
                        session=None,
                        obj=None,
                        **kwargs):
        session = session if session else self._getSession()
        pl = self._retrieve(session, objId, self._payloadFilter, obj=obj)
        pl.filt.update(self._payloadFilter)
        pl.filt[plIdKey] = objId
        VortexFactory.sendVortexMsg(pl.toVortexMsg(), destVortexUuid=vortexUuid)

    def preProcess(self, payload, vortextUuid, **kwargs):
        pass

    def postProcess(self, action, payloadFilt, vortextUuid):
        pass


class OrmCrudHandlerInThread(OrmCrudHandler):
    def _process(self, payload, **kwargs):
        return deferToThread(OrmCrudHandler.process, self, payload, **kwargs)

    def sendModelUpdate(self, objId, **kwargs):
        return deferToThread(OrmCrudHandler.sendModelUpdate, self, objId, **kwargs)



import json
import logging
from typing import Annotated, List

from twisted.trial import unittest

from vortex.Tuple import (
    Tuple,
    TupleField,
    addTupleType,
    IntTupleFieldValidator,
    TupleFieldValidatorABC,
)
from vortex.restful.RestfulHttpClient import RestfulHttpClient
from vortex.restful.RestfulResource import HTTP_REQUEST

logger = logging.getLogger(__name__)


@addTupleType
class TestTuple(Tuple):
    __tupleType__ = "TestTuple"
    id: int = TupleField()
    count: Annotated[int, IntTupleFieldValidator(0, 100)] = TupleField()
    text: str = TupleField()


@addTupleType
class ResponseTuple(Tuple):
    __tupleType__ = "ResponseTuple"
    data: str = TupleField(comment="request in string")
    json: TestTuple = TupleField(comment="request converted to jsonDict")


class RestfulClientTest(unittest.TestCase):
    def setUp(self):
        self.tuple = TestTuple()
        self.tuple.id = 10
        self.tuple.count = 50
        self.tuple.text = "test"

    def testHttpPost(self):
        client = RestfulHttpClient(
            "http://httpbin.org/anything",
            method=HTTP_REQUEST.POST,
            postTuple=self.tuple,
            ResponseTuple=ResponseTuple,
        )
        d = client.run()
        d.addCallback(self._checkPostJSON)
        # d.addCallback(self.print)
        return d

    def _checkPostJSON(self, tuple_):
        self.assertTrue(isinstance(tuple_, ResponseTuple))
        self.assertEqual(tuple_.data, json.dumps(self.tuple.tupleToRestfulJsonDict()))
        self.assertEqual(tuple_.json.id, self.tuple.id)
        self.assertEqual(tuple_.json.text, self.tuple.text)



"""
 * Created by Synerty Pty Ltd
 *
 * This software is open source, the MIT license applies.
 *
 * Website : http://www.synerty.com
 * Support : support@synerty.com
"""
import gc
import os
import tracemalloc
from collections import deque
from random import random
from typing import Optional

import psutil
from twisted.internet import reactor
from twisted.internet.defer import inlineCallbacks, Deferred
from twisted.python.failure import Failure
from twisted.trial import unittest

from vortex.Payload import Payload
from vortex.PayloadEndpoint import PayloadEndpoint
from vortex.VortexClientTcp import VortexClientTcp
from vortex.VortexServer import VortexServer
from vortex.VortexServerTcp import VortexTcpServerFactory


class MemoryCheckerTestMixin:

    def _memMark(self):
        self._process = psutil.Process(os.getpid())
        unreachable = gc.collect()
        self._initialMem = self._process.memory_info().rss
        self._initialUnreachable = unreachable
        self._lastMem = self._initialMem

    def _memCheck(self, grace=0):
        unreachable = gc.collect()
        self.assertLessEqual(unreachable, self._initialUnreachable,
                             "The garbage collector couldn't release everything")
        self._memPrintIncrease()
        memNow = self._process.memory_info().rss
        self.assertLessEqual(memNow, self._initialMem + grace)

    def _memPrintIncrease(self, printOnlyOnGrowth=False):
        gc.collect()
        memNow = self._process.memory_info().rss
        thisDiff = memNow - self._initialMem
        lastDiff = self._lastMem - self._initialMem
        growth = thisDiff - lastDiff
        self._lastMem = memNow
        if not printOnlyOnGrowth or growth:
            print("Memory growth is %s, total growth %s, total %s" % (
                "{:,d}".format(growth),
                "{:,d}".format(thisDiff),
                "{:,d}".format(memNow)))

    def _memCheckFunction(self, callable, loopCount, *args, **kwargs):
        assert loopCount >= 10, "We need at least 10 loops to check this"
        # Give the memory 5 loops to warmed up, it shouldn't need more than 2
        startMemIndex = 5
        for index in range(loopCount):
            if index == startMemIndex:  # Mark the second loop only
                self._memMark()

            callable(*args, **kwargs)

            if startMemIndex <= index:
                self._memPrintIncrease(printOnlyOnGrowth=True)

        self._memCheck(grace=0)


class VortexTcpConnectTestMixin:

    @inlineCallbacks
    def _connect(self):
        port = 20000 + int(random() * 10000)

        # Create the server
        self._vortexServer = VortexServer('test server')
        vortexTcpServerFactory = VortexTcpServerFactory(self._vortexServer)
        self._listenPort = reactor.listenTCP(port, vortexTcpServerFactory)

        # Create the client
        self._vortexClient = VortexClientTcp('test client')
        yield self._vortexClient.connect('127.0.0.1', port)

    def _disconnect(self):
        self._listenPort.stopListening()
        del self._listenPort

        self._vortexServer.shutdown()
        del self._vortexServer

        self._vortexClient.close()
        del self._vortexClient


class VortexSendReceiveTestMixin:
    # Setup a queue state
    _loopbackFilt = {'key': 'unittest'}

    class _State:
        def __init__(self, testSelf, sendFromServerOnly):
            self.testSelf = testSelf
            self.sendFromServerOnly = sendFromServerOnly

            self.dataSizeStats = []
            self.vortexSizeStats = []

            self.dataQueue = deque()
            self.dataQueueEmptyDeferred = Deferred()
            self.totalSent = 0
            self.totalReceived = 0
            self.totalMessagesSent = 0
            self.totalMessagesReceived = 0
            self.totalMessagesOutOfOrder = 0
            self.highestOutOfOrderIndex = 0

            self.totalSentFromClient = 0
            self.totalSentFromServer = 0

        def vortex(self, dataLen: int):
            if not self.sendFromServerOnly and random() < 0.5:
                self.totalSentFromClient += dataLen
                return self.testSelf._vortexClient

            self.totalSentFromServer += dataLen
            return self.testSelf._vortexServer

    class _Checker:

        def __init__(self, testSelf, state, printStatusEveryXMessage):
            self.endpoint = None
            self.testSelf = testSelf
            self.state = state
            self.printStatusEveryXMessage = printStatusEveryXMessage

        def setEndpoint(self, endpoint):
            self.endpoint = endpoint

        # Setup the receiver of the data
        @inlineCallbacks
        def process(self, payloadEnvelope, *args, **kwargs):
            try:
                payload = yield payloadEnvelope.decodePayloadDefer()
                dataReceived = payload.tuples[0]
                self.state.totalReceived += len(dataReceived)
                self.state.totalMessagesReceived += 1

                dataSent = None
                for index, item in enumerate(self.state.dataQueue):
                    if item == dataReceived:
                        dataSent = dataReceived
                        if index:
                            self.state.totalMessagesOutOfOrder += 1
                            self.state.highestOutOfOrderIndex = \
                                max(self.state.highestOutOfOrderIndex, index)
                            self.state.dataQueue.remove(item)

                        else:
                            self.state.dataQueue.popleft()

                        break

                if not (self.state.totalMessagesReceived % self.printStatusEveryXMessage):
                    print("Received message %s, this %s, total %s"
                          % (self.state.totalMessagesReceived,
                             "{:,d}".format(len(dataReceived)),
                             "{:,d}".format(self.state.totalReceived)))

                self.testSelf.assertEqual(dataReceived, dataSent)

                if not self.state.dataQueue:
                    self.state.dataQueueEmptyDeferred.callback(True)

            except Exception as e:
                self.endpoint.shutdown()
                self.state.dataQueueEmptyDeferred.errback(Failure(e))

    @inlineCallbacks
    def _vortexTestTcpServerClient(self, printStatusEveryXMessage: Optional[int] = 1000,
                                   maxMessageSizeBytes: Optional[int] = None,
                                   exactMessageSizeBytes: Optional[int] = None,
                                   totalBytesToSend: Optional[int] = None,
                                   totalMessagesToSend: Optional[int] = None,
                                   payloadCompression=9,
                                   sendFromServerOnly=False):
        assert totalBytesToSend or totalMessagesToSend, "We must have a total to send"
        assert maxMessageSizeBytes or exactMessageSizeBytes, "We must have a msg size"

        state = self._State(self, sendFromServerOnly)
        checker = self._Checker(self, state, printStatusEveryXMessage)
        endpoint = PayloadEndpoint(self._loopbackFilt, checker.process)
        checker.setEndpoint(endpoint)

        # Make random chunks of data
        def makeData():
            size = random() * maxMessageSizeBytes \
                if exactMessageSizeBytes is None \
                else exactMessageSizeBytes

            packet = str(random())
            while len(packet) < size:
                packet += str(random())

            if exactMessageSizeBytes is not None:
                packet = packet[:exactMessageSizeBytes]

            return packet

        def check():
            if totalMessagesToSend and state.totalMessagesSent < totalMessagesToSend:
                return True

            if totalBytesToSend and state.totalSent < totalBytesToSend:
                return True

            return False

        # Send the data
        while check():
            data = makeData()
            state.dataQueue.append(data)
            state.totalSent += len(data)
            state.totalMessagesSent += 1

            vortexMsg = yield Payload(self._loopbackFilt, data) \
                .makePayloadEnvelopeVortexMsgDefer(compressionLevel=payloadCompression)

            state.dataSizeStats.append(len(data))
            state.vortexSizeStats.append(len(vortexMsg))

            # We could send this from either the vortexClient or vortexServer
            # We only have one PayloadIO that the PayloadEndpoint binds to anyway
            state.vortex(len(data)).sendVortexMsg(vortexMsg)

            if not (state.totalMessagesSent % printStatusEveryXMessage):
                print("Sent     message %s, this %s, total %s"
                      % (state.totalMessagesSent,
                         "{:,d}".format(len(data)),
                         "{:,d}".format(state.totalSent)))

        # Wait for all the sending to complete
        yield state.dataQueueEmptyDeferred

        print("%s messages were out of order, the most out of order was %s"
              % (state.totalMessagesOutOfOrder, state.highestOutOfOrderIndex))

        print("Sent %s from the vortex client"
              % "{:,d}".format(state.totalSentFromClient))

        print("Sent %s from the vortex server"
              % "{:,d}".format(state.totalSentFromServer))

        print("Data     : count %s, total size %s, max size %s, min size %s, average %s"
              % (state.totalMessagesSent,
                 "{:,d}".format(state.totalSent),
                 "{:,d}".format(max(state.dataSizeStats)),
                 "{:,d}".format(min(state.dataSizeStats)),
                 "{:,d}".format(int(state.totalSent / state.totalMessagesSent))))

        totalVortexMsgs = sum(state.vortexSizeStats)
        print("VortexMsg: count %s, total size %s, max size %s, min size %s, average %s"
              % (state.totalMessagesSent,
                 "{:,d}".format(totalVortexMsgs),
                 "{:,d}".format(max(state.vortexSizeStats)),
                 "{:,d}".format(min(state.vortexSizeStats)),
                 "{:,d}".format(int(totalVortexMsgs / state.totalMessagesSent))))

        # Run our checks
        self.assertEqual(state.totalSent, state.totalReceived)
        self.assertFalse(len(state.dataQueue))

        # Cleanup
        endpoint.shutdown()
        del checker
        del state
        del endpoint

        gc.collect()


class VortexTcpConnectTest(unittest.TestCase,
                           MemoryCheckerTestMixin,
                           VortexTcpConnectTestMixin,
                           VortexSendReceiveTestMixin):
    @inlineCallbacks
    def __vortexReconnect(self, count):
        self._memMark()
        for x in range(count):
            print("Reconnecting #%s" % x)
            yield self._connect()
            self._disconnect()
            self._memPrintIncrease()
        self._memPrintIncrease()

    @inlineCallbacks
    def test_vortexReconnect100(self):
        yield self.__vortexReconnect(100)

    @inlineCallbacks
    def __vortexReconnect_with_data(self, count):
        self._memMark()
        for x in range(count):
            print("Reconnecting #%s" % x)
            yield self._connect()
            yield self._vortexTestTcpServerClient(totalMessagesToSend=10,
                                                  exactMessageSizeBytes=100 * 1024)
            self._disconnect()
            self._memPrintIncrease()
        self._memPrintIncrease()

    @inlineCallbacks
    def test_vortexReconnect100_with_data(self):
        yield self.__vortexReconnect_with_data(100)


class VortexTcpMemoryLeakTest(unittest.TestCase,
                              MemoryCheckerTestMixin,
                              VortexTcpConnectTestMixin,
                              VortexSendReceiveTestMixin):

    @inlineCallbacks
    def __vortexSendMsgCountMsgSize(self, count, size, compression=9):
        self._memMark()
        yield self._connect()
        yield self._vortexTestTcpServerClient(totalMessagesToSend=count,
                                              printStatusEveryXMessage=1000,
                                              exactMessageSizeBytes=size,
                                              payloadCompression=compression)
        self._disconnect()
        self._memPrintIncrease()

    @inlineCallbacks
    def test_1vortexSend1_1kb_compression1(self):
        yield self.__vortexSendMsgCountMsgSize(1, 1 * 1024, 1)

    @inlineCallbacks
    def test_1vortexSend1_300kb_compression1(self):
        yield self.__vortexSendMsgCountMsgSize(1, 300 * 1024, 1)

    @inlineCallbacks
    def test_1vortexSend10000_1kb_compression1(self):
        yield self.__vortexSendMsgCountMsgSize(10000, 1 * 1024, 1)

    @inlineCallbacks
    def test_1vortexSend10000_100kb_compression1(self):
        yield self.__vortexSendMsgCountMsgSize(10000, 100 * 1024, 1)

    @inlineCallbacks
    def test_1vortexSend10000_1kb_compression9(self):
        yield self.__vortexSendMsgCountMsgSize(10000, 1 * 1024, 9)

    @inlineCallbacks
    def test_1vortexSend10000_100kb_compression9(self):
        yield self.__vortexSendMsgCountMsgSize(10000, 100 * 1024, 9)

    @inlineCallbacks
    def __vortexSendTotalSendMsgSize(self, totalSend, msgSize,
                                     compression=9, fromServer=False):
        self._memMark()
        yield self._connect()
        yield self._vortexTestTcpServerClient(totalBytesToSend=totalSend,
                                              printStatusEveryXMessage=1000,
                                              maxMessageSizeBytes=msgSize,
                                              sendFromServerOnly=fromServer,
                                              payloadCompression=compression)
        self._disconnect()
        self._memPrintIncrease()

    @inlineCallbacks
    def test_2vortexSend1mb_1kb(self):
        yield self.__vortexSendTotalSendMsgSize(1024 ** 2, 1 * 1024)

    @inlineCallbacks
    def test_2vortexSend1mb_100kb(self):
        yield self.__vortexSendTotalSendMsgSize(1024 ** 2, 100 * 1024)

    @inlineCallbacks
    def test_3vortexSend1mb_X10_yes_reconnect(self):
        mem = MemoryCheckerTestMixin()
        mem._memMark()

        print(" ===== START LOOP ===== ")
        for i in range(10):
            print(" ===== Running iteration %s ===== " % i)
            self._memMark()
            yield self._connect()
            yield self._vortexTestTcpServerClient(totalBytesToSend=1024 ** 2,
                                                  printStatusEveryXMessage=1000,
                                                  maxMessageSizeBytes=100 * 1024)
            self._disconnect()
            self._memPrintIncrease()

        print(" ===== END LOOP ===== ")
        mem._memPrintIncrease()

    @inlineCallbacks
    def test_3vortexSend1mb_X10_no_reconnect(self):
        mem = MemoryCheckerTestMixin()
        mem._memMark()
        # Initialise the vortex
        yield self._connect()

        print(" ===== START LOOP ===== ")
        for i in range(10):
            print(" ===== Running iteration %s ===== " % i)
            # Setup the parameters for the test
            self._memMark()
            yield self._vortexTestTcpServerClient(totalBytesToSend=1024 ** 2,
                                                  printStatusEveryXMessage=1000,
                                                  maxMessageSizeBytes=1 * 1024)
            self._memPrintIncrease()

        self._disconnect()
        print(" ===== END LOOP ===== ")
        mem._memPrintIncrease()

    @inlineCallbacks
    def test_4vortexSend10mb(self):
        yield self.__vortexSendTotalSendMsgSize(1024 ** 2 * 10, 100 * 1024)

    @inlineCallbacks
    def test_5vortexSend100mb(self):
        yield self.__vortexSendTotalSendMsgSize(1024 ** 2 * 100, 1024 ** 2)

    @inlineCallbacks
    def test_6vortexSend500mb(self):
        yield self.__vortexSendTotalSendMsgSize(1024 ** 2 * 500, 1024 ** 2)


class VortexTcpMemoryLeakLargeMessagesTest(unittest.TestCase,
                                           MemoryCheckerTestMixin,
                                           VortexTcpConnectTestMixin,
                                           VortexSendReceiveTestMixin):
    # Increase the timeout to 15 minutes
    timeout = 30 * 60

    _patchedProcessData_dataReceivedLen = 0

    def _processData_patch(self):
        VortexTcpMemoryLeakLargeMessagesTest._patchedProcessData_dataReceivedLen += len(
            self._data)

        # Just drop the data and continue
        self._data = b''

    @inlineCallbacks
    def __vortexSend1g_oneway_maxmsg(self, msgSize, totalSize=1024 ** 2 * 500):
        tracemalloc.start(5)
        self._memMark()
        yield self._connect()
        yield self._vortexTestTcpServerClient(totalBytesToSend=totalSize,
                                              printStatusEveryXMessage=1000,
                                              exactMessageSizeBytes=msgSize,
                                              sendFromServerOnly=True,
                                              payloadCompression=0)

        self._disconnect()

        snapshot = tracemalloc.take_snapshot()
        top_stats = snapshot.statistics('lineno')

        print("[ Top 10 ]")
        for stat in top_stats[:10]:
            print(stat)

        tracemalloc.stop()

        for _ in range(5):
            d = Deferred()
            reactor.callLater(2, d.callback, True)
            print("Sleeping 2 seconds")
            yield d
            self._memPrintIncrease()


    @inlineCallbacks
    def test_1vortexSend50mb_oneway_msg10kb(self):
        yield self.__vortexSend1g_oneway_maxmsg(10 * 1024, totalSize=1024 ** 2 * 50)

    @inlineCallbacks
    def test_1vortexSend50mb_oneway_msg100kb(self):
        yield self.__vortexSend1g_oneway_maxmsg(100 * 1024, totalSize=1024 ** 2 * 50)

    @inlineCallbacks
    def test_1vortexSend50mb_oneway_msg1mb(self):
        yield self.__vortexSend1g_oneway_maxmsg(1 * 1024 ** 2, totalSize=1024 ** 2 * 50)

    @inlineCallbacks
    def test_1vortexSend50mb_oneway_msg3mb(self):
        yield self.__vortexSend1g_oneway_maxmsg(3 * 1024 ** 2, totalSize=1024 ** 2 * 50)

    @inlineCallbacks
    def test_1vortexSend50mb_oneway_msg10mb(self):
        yield self.__vortexSend1g_oneway_maxmsg(10 * 1024 ** 2, totalSize=1024 ** 2 * 50)

    @inlineCallbacks
    def test_1vortexSend50mb_oneway_msg50mb(self):
        yield self.__vortexSend1g_oneway_maxmsg(50 * 1024 ** 2, totalSize=1024 ** 2 * 50)

    @inlineCallbacks
    def test_1vortexSend500mb_oneway_msg10mb(self):
        yield self.__vortexSend1g_oneway_maxmsg(10 * 1024 ** 2, totalSize=500 * 1024 ** 2)


class VortexTcpMemoryLeakLargeTransfersTest(unittest.TestCase,
                                            MemoryCheckerTestMixin,
                                            VortexTcpConnectTestMixin,
                                            VortexSendReceiveTestMixin):

    @inlineCallbacks
    def __vortexSendTest(self, total, compression):
        self._memMark()
        yield self._connect()
        yield self._vortexTestTcpServerClient(totalBytesToSend=total,
                                              printStatusEveryXMessage=10000,
                                              maxMessageSizeBytes=10 * 1024 ** 2,
                                              payloadCompression=compression)
        self._disconnect()
        self._memPrintIncrease()

    @inlineCallbacks
    def test_2vortexSend1gb_compression1(self):
        yield self.__vortexSendTest(1024 ** 3, compression=1)

    @inlineCallbacks
    def test_3vortexSend3gb_compression1(self):
        yield self.__vortexSendTest(1024 ** 3 * 3, compression=1)



"""
 * Created by Synerty Pty Ltd
 *
 * This software is open source, the MIT license applies.
 *
 * Website : http://www.synerty.com
 * Support : support@synerty.com
"""
import json
from typing import Any, Dict, Optional

from vortex.Tuple import Tuple, addTupleType, TupleField


@addTupleType
class TupleSelector(Tuple):
    __tupleType__ = "vortex.TupleSelector"
    __slots__ = ["name", "selector"]

    # name: str = TupleField(comment="The tuple name this selector is for")
    # selector: Dict[str, Any] = TupleField(comment="The values to select")

    def __init__(self, name: Optional[str] = None,
                 selector: Optional[Dict] = None) -> None:
        Tuple.__init__(self)
        self.name = name
        self.selector = selector if selector else {}

    def __eq__(x, y):
        return x.toJsonStr() == y.toJsonStr()

    def __hash__(self):
        return hash(self.toJsonStr())

    def toJsonStr(self) -> str:
        """ To Json Str

        This method dumps the c{TupleSelector} data to a json string.

        It sorts the dict keys and
        """
        fieldJsonDict = self.toJsonField(self.selector)
        return json.dumps({'name': self.name,
                           'selector': fieldJsonDict}, sort_keys=True)

    @classmethod
    def fromJsonStr(self, jsonStr: str) -> "TupleSelector":
        """ From Json Str

        This method creates a new c{TupleSelector} from the ordered json string dumped
        from .toJsonStr

        """
        data = json.loads(jsonStr)
        newTs = TupleSelector(name=data["name"], selector={})
        newTs.selector = newTs.fromJsonField(data["selector"])
        return newTs

    def __repr__(self):
        return self.toJsonStr()



import logging
from functools import wraps
from typing import Union

import twisted
from twisted.internet import reactor
from twisted.internet.defer import (
    maybeDeferred,
    Deferred,
    inlineCallbacks,
    ensureDeferred,
)
from twisted.internet.threads import deferToThread

logger = logging.getLogger(__name__)


def callMethodLater(func):
    def wrapper(*args, **kwargs):
        return reactor.callLater(0, func, *args, **kwargs)

    return wrapper


def nonConcurrentMethod(method):
    def call(*args, **kwargs):
        if call.running:
            return
        call.running = True
        try:
            return method(*args, **kwargs)
        finally:
            call.running = False

    call.running = False

    return call


def maybeDeferredWrap(funcToWrap):
    """Maybe Deferred Wrap

    A decorator that ensures a function will return a deferred.

    """

    def func(*args, **kwargs):
        return maybeDeferred(funcToWrap, *args, **kwargs)

    return func


def ensureDeferredWrap(funcToWrap):
    """Ensured Deferred Wrap

    A decorator that converts asyncio functions to ones that
     return Twisted Deferred instead.

    This keeps all the async processing in Twisteds reactive design while providing
     support of the await/async syntax (and asyncio processing)


    """

    def func(*args, **kwargs):
        return ensureDeferred(funcToWrap(*args, **kwargs))

    return func


def vortexLogFailure(failure, loggerArg, consumeError=False, successValue=True):
    try:
        if not hasattr(failure, "_vortexLogged"):
            if failure.getTraceback():
                loggerArg.error(failure.getTraceback())
            failure._vortexFailureLogged = True
        return successValue if consumeError else failure
    except Exception as e:
        logger.exception(e)


def vortexLogAndConsumeFailure(failure, loggerArg, successValue=True):
    return vortexLogFailure(
        failure, loggerArg, consumeError=True, successValue=successValue
    )


printFailure = vortexLogFailure


def vortexInlineCallbacksLogAndConsumeFailure(loggerArg):
    """Vortex InlineCallbacks Log and Consume Failure

    This is exactly the same as @inlineCallbacks decorator, except it will log
    and consume any deferred failures that are thrown.
    """

    def wrapper(funcToWrap):
        funcIcb = inlineCallbacks(funcToWrap)

        def called(*args, **kwargs):
            d = funcIcb(*args, **kwargs)
            d.addErrback(vortexLogAndConsumeFailure, loggerArg)
            return d

        return called

    return wrapper


def deferToThreadWrapWithLogger(logger, consumeError=False, checkMainThread=True):
    """Defer To Thread Wrap With Logger

    This method is a decorator used to send blocking methods to threads easily.

    :param logger: An instance from logging.getLogger(..)
    :param consumeError: Should errors occuring in the thread be consumed or errback'd
    :param checkMainThread: Don't use this parameter, it's only needed during reactor
                startup when the reactors thread isn't properly set and the check
                fails.

    :return: A deferred.

    Usage: ::

            import logging
            logger = logging.getLogger(__name__)
            @deferToThreadWrapWithLogger(logger)
            def myFunction(arg1, kw=True):
                pass


    """
    assert isinstance(
        logger, logging.Logger
    ), """Usage:
    import logging
    logger = logging.getLogger(__name__)
    @deferToThreadWrapWithLogger(logger)
    def myFunction(arg1, kw=True):
        pass
    """

    def wrapper(funcToWrap) -> Deferred:
        def func(*args, **kwargs):
            if not twisted.python.threadable.isInIOThread() and checkMainThread:
                raise Exception(
                    "Deferring to a thread can only be done from the main thread"
                )

            d = deferToThread(funcToWrap, *args, **kwargs)
            d.addErrback(vortexLogFailure, logger, consumeError=consumeError)
            return d

        return func

    return wrapper


def isMainThread():
    return twisted.python.threadable.isInIOThread()


def noMainThread():
    if isMainThread():
        raise Exception("Blocking operations shouldn't occur in the reactors thread.")


def yesMainThread():
    if not isMainThread():
        raise Exception("Async operations must occur in the reactors main thread.")



"""
 * Created by Synerty Pty Ltd
 *
 * This software is open source, the MIT license applies.
 *
 * Website : http://www.synerty.com
 * Support : support@synerty.com
"""
import logging
import uuid
from typing import Optional, Union, Dict, List, Any
from weakref import WeakValueDictionary

from twisted.internet import task, reactor
from twisted.internet.defer import Deferred, inlineCallbacks
from twisted.python.components import registerAdapter
from twisted.web.server import Session
from zope.interface import Interface, Attribute
from zope.interface.declarations import implementer

from vortex.DeferUtil import vortexLogFailure, isMainThread
from vortex.PayloadEnvelope import VortexMsgList, PayloadEnvelope
from vortex.PayloadFilterKeys import rapuiServerEcho
from vortex.PayloadIO import PayloadIO
from vortex.PayloadPriority import DEFAULT_PRIORITY
from vortex.VortexABC import VortexABC, VortexInfo

logger = logging.getLogger(__name__)

''' ---------------------------------------------------------------------------
VortexServer
'''

HEART_BEAT_PERIOD = 5.0
HEART_BEAT_TIMEOUT = 35.0


class VortexServer(VortexABC):
    ''' VortexServer
    The static instance of the controller
    '''

    def __init__(self, name: str) -> None:
        # Simple initialisations up the top
        self._name = name
        self._uuid = str(uuid.uuid1())
        self._shutdown = False

        # Store all our sessions
        self._httpSessionsBySessionUuid: WeakValueDictionary = WeakValueDictionary()
        self._connectionByVortexUuid: Dict[str, Any] = {}

    def name(self):
        return self._name

    def uuid(self):
        return self._uuid

    @property
    def localVortexInfo(self) -> VortexInfo:
        return VortexInfo(name=self._name,
                          uuid=self._uuid)

    @property
    def remoteVortexInfo(self) -> List[VortexInfo]:
        vortexInfos = []

        for conn in self._connectionByVortexUuid.values():
            vortexInfos.append(
                VortexInfo(name=conn.remoteVortexName, uuid=conn.remoteVortexUuid))

        return vortexInfos

    def isShutdown(self):
        return self._shutdown

    def shutdown(self):
        self._shutdown = True

        for sess in list(self._httpSessionsBySessionUuid.values()):
            sess.expire()

        for conn in list(self._connectionByVortexUuid.values()):
            conn.close()

    def vortexUuids(self):
        return list(self._connectionByVortexUuid.keys())

    def isVortexAlive(self, vortexUuid):
        return vortexUuid in self._connectionByVortexUuid

    def vortexClientIpPort(self, vortexUuid):
        connection = self._connectionByVortexUuid.get(vortexUuid)
        if not connection:
            return None

        return "%s:%s" % (connection.ip, connection.port)

    def connectionOpened(self, httpSession, vortexConnection):
        # print "VortexServer - connectionOpened"

        vortexUuid = vortexConnection.remoteVortexUuid
        assert vortexUuid

        if httpSession:
            # If this is a new session, Make sure we have an expire callback on it
            if httpSession.uid not in self._httpSessionsBySessionUuid:
                self._httpSessionsBySessionUuid[httpSession.uid] = httpSession
                httpSession.notifyOnExpire(lambda: self._sessionExpired(httpSession.uid))

            # Update the connection dict in the sessions object
            httpSessionConns = VortexSessionI(httpSession).connections
            httpSessionConns[vortexUuid] = vortexConnection

        # Get the old connection if it exists
        if vortexUuid in self._connectionByVortexUuid:
            self._connectionByVortexUuid[vortexUuid].close()

        # Update the _connectionsByvortexUuid
        self._connectionByVortexUuid[vortexUuid] = vortexConnection

    def connectionClosed(self, conn):
        # print "VortexServer - connectionClosed"

        vortexUuid = conn.remoteVortexUuid

        if conn.httpSessionUuid in self._httpSessionsBySessionUuid:
            session = self._httpSessionsBySessionUuid[conn.httpSessionUuid]
            assert session

            # cleanup _sessionsBySessionUuid
            conns = VortexSessionI(session).connections
            if conns[vortexUuid] == conn:
                del conns[vortexUuid]

        # cleanup _connectionsByvortexUuid
        if conn.remoteVortexUuid in self._connectionByVortexUuid:
            if self._connectionByVortexUuid[conn.remoteVortexUuid] == conn:
                del self._connectionByVortexUuid[conn.remoteVortexUuid]

    def _sessionExpired(self, httpSessionUuid):
        logger.debug(
            "VortexServer - _sessionExpired, Session %s has expired" % httpSessionUuid)

        # cleanup _sessionsBySessionUuid
        del self._httpSessionsBySessionUuid[httpSessionUuid]

        # cleanup _connectionsByvortexUuid
        for vortexUuid, conn in list(self._connectionByVortexUuid.items()):
            if conn.httpSessionUuid == httpSessionUuid:
                del self._connectionByVortexUuid[vortexUuid]

    def payloadReveived(self, httpSession, vortexUuid, vortexName, payload):
        # print "VortexServer - payloadReveived"

        if rapuiServerEcho in payload.filt:
            payload.filt.pop(rapuiServerEcho)
            self.sendVortexMsg(payload.toVortexMsg(), vortexUuid)

        def sendResponse(vortexMsg: bytes, priority: int = DEFAULT_PRIORITY) -> Deferred:
            """ Send Back

            Sends a response back to where this payload come from.

            """
            return self.sendVortexMsg(vortexMsg, vortexUuid=vortexUuid, priority=priority)

        PayloadIO().process(payload,
                            vortexUuid=vortexUuid,
                            vortexName=vortexName,
                            httpSession=httpSession,
                            sendResponse=sendResponse
                            )

    def sendVortexMsg(self,
                      vortexMsgs: Union[VortexMsgList, bytes, None] = None,
                      vortexUuid: Optional[str] = None,
                      priority: int = DEFAULT_PRIORITY):
        """ Send Vortex Msg

        Sends the vortex message to any conencted clients with vortexUuid.
        Or broadcast it to all connected vortex clients if it's None

        :param vortexMsgs: The vortex message to send
        :param vortexUuid: The vortexUuid of the client to send to.
        """
        if vortexMsgs is None:
            vortexMsgs = [PayloadEnvelope().toVortexMsg()]

        if isMainThread():
            return self._sendVortexMsgLater(vortexMsgs, vortexUuid=vortexUuid,
                                            priority=priority)

        return task.deferLater(reactor, 0, self._sendVortexMsgLater, vortexMsgs,
                               vortexUuid=vortexUuid, priority=priority)

    @inlineCallbacks
    def _sendVortexMsgLater(self,
                            vortexMsgs: Union[VortexMsgList, bytes],
                            vortexUuid: Optional[str],
                            priority: int):
        """ Send the message.

        Send it later,
        This also means it doesn't matter what thread this is called from

        """
        yield None

        if not isinstance(vortexMsgs, list):
            vortexMsgs = [vortexMsgs]

        # Deliver locally
        if vortexUuid == self._uuid:
            for vortexMsg in vortexMsgs:
                def sendResponse(vortexMsg_, priority_=DEFAULT_PRIORITY):
                    self._sendVortexMsgLater(vortexMsg_, self._uuid, priority_)

                def cb(payloadEnvelope: PayloadEnvelope) -> None:
                    PayloadIO().process(
                        payloadEnvelope,
                        vortexUuid=self._uuid,
                        vortexName=self._name,
                        httpSession=None,
                        sendResponse=sendResponse)

                d = PayloadEnvelope.fromVortexMsgDefer(vortexMsg)
                d.addCallback(cb)
                d.addErrback(vortexLogFailure, logger, consumeError=True)

            return

        from vortex.VortexServerConnection import VortexServerConnection
        conns: List[VortexServerConnection] = []
        if vortexUuid is None:
            conns = list(self._connectionByVortexUuid.values())
        elif vortexUuid in self._connectionByVortexUuid:
            conns.append(self._connectionByVortexUuid[vortexUuid])

        for conn in conns:
            for vortexMsg in vortexMsgs:
                conn.write(vortexMsg, priority)

        return True


''' ---------------------------------------------------------------------------
VortexServer Session
'''


class VortexSessionI(Interface):
    connections: Dict = Attribute(
        "VortexServer connections for this session, by window uuid")


@implementer(VortexSessionI)
class VortexSession(object):
    def __init__(self, session):
        session.sessionTimeout = HEART_BEAT_TIMEOUT
        self.connections = {}


registerAdapter(VortexSession, Session, VortexSessionI)



"""
 * Created by Synerty Pty Ltd
 *
 * This software is open source, the MIT license applies.
 *
 * Website : http://www.synerty.com
 * Support : support@synerty.com
"""
import gc

from twisted.trial import unittest

from vortex.Payload import Payload
from vortex.PayloadFilterKeys import rapuiClientEcho
from vortex.Tuple import TupleHash
from vortex.test.VortexTcpMemoryLeakPayloadTest import MemoryCheckerTestMixin


def makeTestPayloadA(tupleDataSize=10 * 1024):
    payload = Payload()
    payload.filt[rapuiClientEcho] = None
    payload.tuples = ['12345678' * int(tupleDataSize / 8)]
    return payload


class PayloadPyTest(unittest.TestCase, MemoryCheckerTestMixin):
    PERF_TEST_COUNT = 10000

    def setUp(self):
        pass

    def tearDown(self):
        pass

    # @profile
    def _makeStr(self, size):
        _ = ['12345678' * int(size / 8)]
        del _
        gc.collect()

    # @profile
    def _payloadEncodeDecode(self, size):
        origPayload = makeTestPayloadA(size)
        origPayload.date = None

        encodedPayload = origPayload.toEncodedPayload()
        payload = Payload().fromEncodedPayload(encodedPayload)

        self.assertEqual(payload.tuples[0], origPayload.tuples[0])

    def testMakeStr_100m_x100(self):
        self._memCheckFunction(self._makeStr, 1000, 100 * 1024 ** 2)

    def testPayloadToFromEncodedPayload_10k_x1000(self):
        self._memCheckFunction(self._payloadEncodeDecode, 1000, 10 * 1024)

    def testPayloadToFromEncodedPayload_100m_x100(self):
        self._memCheckFunction(self._payloadEncodeDecode, 100, 10 * 1024 ** 2)

    def testPayloadToFromEncodedPayload_1m_x1000(self):
        self._memCheckFunction(self._payloadEncodeDecode, 1000, 1 * 1024 ** 2)

    def testPayloadToFromEncodedPayload10mb(self):
        # Create Payload
        origPayload = makeTestPayloadA()
        origPayload.tuples = ['1234567890' * 1024 ** 2]

        encodedPayload = origPayload.toEncodedPayload()
        payload = Payload().fromEncodedPayload(encodedPayload)

        self.assertEqual(payload.tuples[0], origPayload.tuples[0])

    def testPayloadToFromEncodedPayload(self):
        # Create Payload
        origPayload = makeTestPayloadA()

        # To VortexServer Message
        encodedPayload = origPayload.toEncodedPayload()
        # print encodedPayload

        payload = Payload().fromEncodedPayload(encodedPayload)

        self.assertEqual(TupleHash(payload), TupleHash(payload))
        self.assertEqual(TupleHash(origPayload), TupleHash(origPayload))

        self.assertEqual(TupleHash(origPayload), TupleHash(payload),
                         'Payload serialise -> deserialize error')

    def testPayloadToFromJson(self):
        # Create Payload
        origPayload = makeTestPayloadA()

        # To JSON
        jsonStr = origPayload._toJson()
        # print encodedPayload
        # print(jsonStr)

        payload = Payload()._fromJson(jsonStr)

        self.assertEqual(TupleHash(payload), TupleHash(payload))
        self.assertEqual(TupleHash(origPayload), TupleHash(origPayload))

        self.assertEqual(TupleHash(origPayload), TupleHash(payload),
                         'Payload serialise -> deserialise error')

    def testPayloadJsonPerformanceSingle(self):
        origPayload = makeTestPayloadA()
        for _ in range(self.PERF_TEST_COUNT):
            origPayload.tuples.append(makeTestPayloadA())
        jsonStr = origPayload._toJson()
        payload = Payload()._fromJson(jsonStr)

    def testPayloadEncodedPayloadPerformanceSingle(self):
        origPayload = makeTestPayloadA()
        for x in range(self.PERF_TEST_COUNT):
            origPayload.tuples.append(makeTestPayloadA())
        encodedPayload = origPayload.toEncodedPayload()
        payload = Payload().fromEncodedPayload(encodedPayload)

    def testPayloadJsonPerformanceMultiple(self):
        origPayload = makeTestPayloadA()
        for _ in range(self.PERF_TEST_COUNT):
            jsonStr = origPayload._toJson()
            payload = Payload()._fromJson(jsonStr)

    def testPayloadEncodedPayloadPerformanceMultiple(self):
        origPayload = makeTestPayloadA()
        for _ in range(self.PERF_TEST_COUNT):
            encodedPayload = origPayload.toEncodedPayload()
            payload = Payload().fromEncodedPayload(encodedPayload)



"""
 * Created by Synerty Pty Ltd
 *
 * This software is open source, the MIT license applies.
 *
 * Website : http://www.synerty.com
 * Support : support@synerty.com
"""
from vortex.test.TestTuple import TestTuple


def makeTestTupleData(count=5):
    tuples = []
    for num in range(count):
        uniStr = "#%s double hyphen :-( — “fancy quotes”" % num
        tuples.append(TestTuple(aInt=num,
                              aBoolTrue=bool(num % 2),
                              aString="This is tuple #%s" % num,
                              aStrWithUnicode=uniStr))

    return tuples



import logging
import typing
from collections import defaultdict
from typing import Union, List, Optional, Dict

from rx.subjects import Subject
from twisted.internet import reactor
from twisted.internet.defer import Deferred, DeferredList, succeed
from twisted.python.failure import Failure
from txwebsocket.txws import WebSocketFactory

from vortex.DeferUtil import yesMainThread
from vortex.PayloadEnvelope import VortexMsgList, PayloadEnvelope
from vortex.PayloadIO import PayloadIO
from vortex.PayloadPriority import DEFAULT_PRIORITY
from vortex.VortexABC import VortexABC
from vortex.VortexClientHttp import VortexClientHttp
from vortex.VortexClientTcp import VortexClientTcp
from vortex.VortexServer import VortexServer
from vortex.VortexServerHttpResource import VortexServerHttpResource
from vortex.VortexServerTcp import VortexTcpServerFactory
from vortex.VortexServerWebsocket import VortexWebsocketServerFactory, \
    VortexWebSocketUpgradeResource, VortexWrappedWebSocketFactory

logger = logging.getLogger(__name__)

broadcast = None

browserVortexName = "browser"


class NoVortexException(Exception):
    """ No Vortex Exception

    This is raised when a remote vortex doesn't exist.

    """


VortexUuidList = List[str]


class VortexFactory:
    __vortexServersByName: Dict[str, List[VortexABC]] = defaultdict(list)
    __vortexClientsByName: Dict[str, List[VortexABC]] = defaultdict(list)

    __vortexStatusChangeSubjectsByName: Dict[str, Subject] = {}

    __isShutdown = False

    def __init__(self):
        raise Exception("Vortex Factory should not be instantiated")

    @classmethod
    def shutdown(cls) -> None:
        cls.__isShutdown = True

    @classmethod
    def _getVortexSendRefs(cls, name=None, uuid=None
                           ) -> List[typing.Tuple[VortexABC, List[str]]]:
        assert name or uuid

        results = []

        # logger.debug("-" * 80)
        for vortex in cls._allVortexes():
            uuids: List[str] = []
            # logger.debug("FROM : %s", vortex.localVortexInfo)
            for remoteVortexInfo in vortex.remoteVortexInfo:
                # logger.debug("        REMOTE : %s", remoteVortexInfo)
                if ((name is None or remoteVortexInfo.name == name)
                        and (uuid is None or remoteVortexInfo.uuid == uuid)):
                    uuids.append(remoteVortexInfo.uuid)

            if uuids:
                results.append((vortex, uuids))

        return results

    @classmethod
    def _allVortexes(cls):
        """ All Vortexes

        :return: A list of all the vortexes, both client and server
        """
        vortexes = []
        for vortexList in cls.__vortexServersByName.values():
            vortexes += vortexList
        for vortexList in cls.__vortexClientsByName.values():
            vortexes += vortexList

        return vortexes

    @classmethod
    def createServer(cls, name: str, rootResource) -> None:
        """ Create Server

        Create a vortex server, VortexServer clients connect to this vortex serer via HTTP(S)

        VortexServer clients will connect and provide their names. This allows the factory to
        abstract away the vortex UUIDs from their names.

        :param name: The name of the local vortex.
        :param rootResource: The resource to add the vortex to. An implementation of
            C{twisted.web.resource.IResource}
        :return: None
        """

        vortexServer = VortexServer(name)
        cls.__vortexServersByName[name].append(vortexServer)

        vortexResource = VortexServerHttpResource(vortexServer)
        rootResource.putChild(b"vortex", vortexResource)

    @classmethod
    def createWebsocketServer(cls, name: str, port: int) -> None:
        """ Create Server

        Create a vortex server, VortexServer clients connect to this vortex serer via HTTP(S)

        VortexServer clients will connect and provide their names. This allows the factory to
        abstract away the vortex UUIDs from their names.

        :param name: The name of the local vortex.
        :param port: The tcp port to listen on
        :return: None
        """

        vortexServer = VortexServer(name)
        cls.__vortexServersByName[name].append(vortexServer)
        vortexWebsocketServerFactory = VortexWebsocketServerFactory(vortexServer)
        reactor.listenTCP(port, WebSocketFactory(vortexWebsocketServerFactory))

    @classmethod
    def createHttpWebsocketServer(cls, name: str, rootResource) -> None:
        """ Create Server

        Create a vortex server, VortexServer clients connect to this vortex serer via HTTP(S)

        VortexServer clients will connect and provide their names. This allows the factory to
        abstract away the vortex UUIDs from their names.

        :param name: The name of the local vortex.
        :param port: The tcp port to listen on
        :return: None
        """

        vortexServer = VortexServer(name)
        cls.__vortexServersByName[name].append(vortexServer)
        vortexWebsocketServerFactory = VortexWebsocketServerFactory(vortexServer)
        websocketFactory = VortexWrappedWebSocketFactory(vortexWebsocketServerFactory)

        websocketResource = VortexWebSocketUpgradeResource(websocketFactory)
        rootResource.putChild(b"vortexws", websocketResource)

    @classmethod
    def createTcpServer(cls, name: str, port: int) -> None:
        """ Create Server

        Create a vortex server, VortexServer clients connect to this vortex serer via HTTP(S)

        VortexServer clients will connect and provide their names. This allows the factory to
        abstract away the vortex UUIDs from their names.

        :param name: The name of the local vortex.
        :param port: The tcp port to listen on
        :return: None
        """

        vortexServer = VortexServer(name)
        cls.__vortexServersByName[name].append(vortexServer)
        vortexTcpServerFactory = VortexTcpServerFactory(vortexServer)
        reactor.listenTCP(port, vortexTcpServerFactory)

    @classmethod
    def createHttpClient(cls, name: str, host: str, port: int) -> Deferred:
        """ Create Client

        Connect to a vortex Server.

        :param name: The name of the local vortex.
        :param host: The hostname of the remote vortex.
        :param port: The port of the remote vortex.

        :return: A deferred from the VortexHttpClient.connect method
        """

        logger.info('Connecting to Peek Server HTTP %s:%s', host, port)

        vortexClient = VortexClientHttp(name)
        cls.__vortexClientsByName[name].append(vortexClient)

        return vortexClient.connect(host, port)

    @classmethod
    def createTcpClient(cls, name: str, host: str, port: int) -> Deferred:
        """ Create Client

        Connect to a vortex Server.

        :param name: The name of the local vortex.
        :param host: The hostname of the remote vortex.
        :param port: The port of the remote vortex.

        :return: A deferred from the VortexTcpClient.connect method
        """

        logger.info('Connecting to Peek Server TCP %s:%s', host, port)

        vortexClient = VortexClientTcp(name)
        cls.__vortexClientsByName[name].append(vortexClient)

        return vortexClient.connect(host, port)

    @classmethod
    def isVortexNameLocal(cls, vortexName: str) -> bool:
        for vortex in cls._allVortexes():
            if vortex.localVortexInfo.name == vortexName:
                return True

        return False

    @classmethod
    def getLocalVortexClients(cls, localVortexName: str) -> List[VortexABC]:
        vortexes: List[VortexABC] = []

        for items in cls.__vortexClientsByName.values():
            vortexes.extend(
                filter(lambda x: x.localVortexInfo.name == localVortexName, items)
            )

        return vortexes

    @classmethod
    def getRemoteVortexUuids(cls) -> List[str]:
        remoteUuids = []

        for vortex in cls._allVortexes():
            for remoteVortexInfo in vortex.remoteVortexInfo:
                remoteUuids.append(remoteVortexInfo.uuid)

        return remoteUuids

    @classmethod
    def getRemoteVortexName(cls) -> List[str]:
        remoteNames = set()

        for vortex in cls._allVortexes():
            for remoteVortexInfo in vortex.remoteVortexInfo:
                remoteNames.add(remoteVortexInfo.name)

        return list(remoteNames)

    @classmethod
    def sendVortexMsg(cls,
                      vortexMsgs: Union[VortexMsgList, bytes],
                      destVortexName: Optional[str] = broadcast,
                      destVortexUuid: Optional[str] = broadcast) -> Deferred:
        """ Send VortexMsg

        Sends a payload to the remote vortex.

        :param vortexMsgs: The vortex message(s) to send to the remote vortex.

        :param destVortexName: The name of the vortex to send the payload to.
            This can be null, If provided, it's used to limit the vortexes the
             message is sent to.

        :param destVortexUuid: The uuid of the vortex to send the payload to,
            This can be null, If provided, it's used to limit the vortexes the
             message is sent to.


        :return: A C{Deferred} which will callback when the message has been sent.
        """

        vortexAndUuids = cls._getVortexSendRefs(destVortexName, destVortexUuid)
        if not vortexAndUuids:
            raise NoVortexException("Can not find vortexes to send message to,"
                                    " name=%s, uuid=%s"
                                    % (destVortexName, destVortexUuid))

        deferreds = []
        for vortex, uuids in vortexAndUuids:
            for uuid in uuids:
                deferreds.append(vortex.sendVortexMsg(vortexMsgs, uuid))

        return DeferredList(deferreds)

    @classmethod
    def sendVortexMsgLocally(cls, vortexMsgs: Union[VortexMsgList, bytes],
                             priority: int = DEFAULT_PRIORITY) -> Deferred:
        """ Send VortexMsg

        Sends a payload to the remote vortex.

        :param vortexMsgs: The vortex message(s) to deliver locally.

        :return: A C{Deferred} which will callback when the message has been delivered.
        """
        yesMainThread()

        vortexUuid = "local"
        vortexName = "local"
        httpSession = "local"
        sendResponse = VortexFactory.sendVortexMsgLocally

        vortexMsgs = [vortexMsgs] if isinstance(vortexMsgs, bytes) else vortexMsgs

        def send(payloadEnvelope: PayloadEnvelope):
            try:
                PayloadIO().process(
                    payloadEnvelope=payloadEnvelope,
                    vortexUuid=vortexUuid,
                    vortexName=vortexName,
                    httpSession=httpSession,
                    sendResponse=sendResponse
                )
                return succeed(True)

            except Exception as e:
                return Failure(e)

        deferreds = []
        for vortexMsg in vortexMsgs:
            d = PayloadEnvelope().fromVortexMsgDefer(vortexMsg)
            d.addCallback(send)
            deferreds.append(d)

        return DeferredList(deferreds)

    @classmethod
    def subscribeToVortexStatusChange(cls, vortexName: str) -> Subject:
        """ Subscribe to Vortex Status Change
        
        Subscribing to the returned observable/subject will provided updates of when the
        vortex goes offline, or online.
        
        .. warning:: This is only implemented for TCP Client vortexes.
        
        :param vortexName: The name of the vortex to subscribe to the status for.
                            This will be the name of the remote vortex.
        
        """
        if not vortexName in cls.__vortexStatusChangeSubjectsByName:
            cls.__vortexStatusChangeSubjectsByName[vortexName] = Subject()
        return cls.__vortexStatusChangeSubjectsByName[vortexName]

    @classmethod
    def _notifyOfVortexStatusChange(cls, vortexName: str, online: bool) -> None:
        if cls.__isShutdown:
            return

        logger.debug("Vortex %s went %s", vortexName, ("online" if online else "offline"))

        if vortexName in cls.__vortexStatusChangeSubjectsByName:
            cls.__vortexStatusChangeSubjectsByName[vortexName].on_next(online)



"""
 * Created by Synerty Pty Ltd
 *
 * This software is open source, the MIT license applies.
 *
 * Website : http://www.synerty.com
 * Support : support@synerty.com
"""
import logging
from datetime import datetime

import pytz
from twisted.internet import task

from .PayloadPriority import DEFAULT_PRIORITY
from .VortexConnectionABC import VortexConnectionABC
from .VortexServer import VortexServer, HEART_BEAT_PERIOD, HEART_BEAT_TIMEOUT
from .VortexWritePushProducer import VortexWritePushProducer

logger = logging.getLogger(name=__name__)


class VortexServerConnection(VortexConnectionABC):
    def __init__(self, vortexServer: VortexServer,
                 remoteVortexUuid: str,
                 remoteVortexName: str,
                 httpSession, transport,
                 addr) -> None:
        VortexConnectionABC.__init__(self,
                                     logger,
                                     vortexServer,
                                     remoteVortexUuid=remoteVortexUuid,
                                     remoteVortexName=remoteVortexName,
                                     httpSessionUuid=httpSession)

        self._lastHeartBeatTime = datetime.now(pytz.utc)
        self._lastHeartBeatCheckTime = datetime.now(pytz.utc)

        self._transport = transport
        self._addr = addr

        # Start our heart beat
        self._beatLoopingCall = task.LoopingCall(self._beat)
        d = self._beatLoopingCall.start(HEART_BEAT_PERIOD, now=False)
        d.addErrback(lambda f: logger.exception(f.value))

        self._producer = None

        # Register the producer if there isn't one already.
        # The websocket server already has one.
        if not self._transport.producer:
            self._producer = VortexWritePushProducer(transport,
                                                     lambda: self.close(),
                                                     remoteVortexName)

            transport.registerProducer(self._producer, True)

    def beatReceived(self):
        self._lastHeartBeatTime = datetime.now(pytz.utc)

    def _beat(self):
        # If we're closed, do nothing
        if self._closed:
            if self._beatLoopingCall.running:
                self._beatLoopingCall.stop()
            return

        beatTimeout = (datetime.now(pytz.utc) - self._lastHeartBeatTime) \
                          .seconds > HEART_BEAT_TIMEOUT

        # If we've been asleep, then make note of that (VM suspended)
        checkTimout = (datetime.now(pytz.utc) - self._lastHeartBeatCheckTime) \
                          .seconds > HEART_BEAT_TIMEOUT

        # Mark that we've just checked it
        self._lastHeartBeatCheckTime = datetime.now(pytz.utc)

        if checkTimout:
            self._lastHeartBeatTime = datetime.now(pytz.utc)
            return

        # If we havn't heard from the client, then close the connection
        if beatTimeout:
            self._beatLoopingCall.stop()
            self.close()
            return

        self._write(b'.', DEFAULT_PRIORITY)

    @property
    def ip(self):
        return self._addr.host

    @property
    def port(self):
        return self._addr.port

    def write(self, payloadVortexStr: bytes, priority: int = DEFAULT_PRIORITY):
        assert not self._closed
        self._write(payloadVortexStr, priority)

    def _write(self, payloadVortexStr: bytes, priority: int):
        if self._producer:
            self._producer.write(payloadVortexStr, priority)
        else:
            self._transport.write(payloadVortexStr)

    def close(self):
        if self._beatLoopingCall.running:
            self._beatLoopingCall.stop()

        self._transport.loseConnection()

    def transportClosed(self):
        self._producer.close()
        VortexConnectionABC.close(self)



import logging
from copy import copy
from typing import List

from rx.subjects import Subject
from twisted.internet.defer import Deferred, inlineCallbacks

from vortex.DeferUtil import vortexLogFailure
from vortex.Payload import Payload
from vortex.PayloadEndpoint import PayloadEndpoint
from vortex.PayloadEnvelope import PayloadEnvelope
from vortex.PayloadResponse import PayloadResponse
from vortex.TupleSelector import TupleSelector
from vortex.VortexFactory import VortexFactory
from vortex.handler.TupleDataObservableCache import TupleDataObservableCache

logger = logging.getLogger(__name__)


class TupleDataObserverClient(TupleDataObservableCache):
    def __init__(self, destVortexName,
                 observableName,
                 additionalFilt=None,
                 observerName="default"):
        """ Constructor

        :param observableName: The name of this observable
        :param additionalFilt: Any additional filter keys that are required
        :param destVortexName: The dest vortex name to send the payloads to

        """
        TupleDataObservableCache.__init__(self)

        self._destVortexName = destVortexName
        self._observableName = observableName
        self._observerName = observerName

        self._sendFilt = dict(name=observableName,
                              observerName=observerName,
                              key="tupleDataObservable")

        self._listenFilt = dict(name=observableName,
                                key="tupleDataObservable")

        if additionalFilt:
            self._sendFilt.update(additionalFilt)
            self._listenFilt.update(additionalFilt)

        self._endpoint = PayloadEndpoint(
            self._listenFilt, self._receivePayload,
            acceptOnlyFromVortex=destVortexName
        )

        # There are no online checks for the vortex
        # isOnlineSub = statusService.isOnline
        # .filter(online= > online == = true)
        # .subscribe(online= > self.vortexOnlineChanged())
        # 
        # self.onDestroyEvent.subscribe(() = > isOnlineSub.unsubscribe())
        VortexFactory \
            .subscribeToVortexStatusChange(destVortexName) \
            .filter(lambda online: online is True) \
            .subscribe(self._vortexOnlineChanged)

        TupleDataObservableCache.start(self)

    def shutdown(self):
        self._endpoint.shutdown()
        TupleDataObservableCache.shutdown(self)

    def pollForTuples(self, tupleSelector: TupleSelector) -> Deferred:
        startFilt = copy(self._sendFilt)
        startFilt.update({"subscribe": False,
                          "tupleSelector": tupleSelector})

        def updateCacheCallback(payloadEnvelope: PayloadEnvelope) -> PayloadEnvelope:
            cache, _ = self._updateCache(payloadEnvelope)
            return payloadEnvelope

        pr = PayloadResponse(payloadEnvelope=PayloadEnvelope(startFilt),
                             destVortexName=self._destVortexName)
        pr.addCallback(updateCacheCallback)
        pr.addCallback(lambda payloadEnvelope: payloadEnvelope.decodePayloadDefer())
        pr.addCallback(lambda payload: payload.tuples)
        return pr

    def subscribeToTupleSelector(self, tupleSelector: TupleSelector) -> Subject:
        cache = self._makeCache(tupleSelector)
        self._tellServerWeWantData([tupleSelector])
        return cache.subject

    def _vortexOnlineChanged(self, *args) -> None:
        self._tellServerWeWantData(self._tupleSelectors())

    @inlineCallbacks
    def _receivePayload(self, payloadEnvelope: PayloadEnvelope, **kwargs):
        # If this message is for a specific observer, and it's not us, then discard it.
        filtObserverName = payloadEnvelope.filt.get("observerName")
        if filtObserverName is not None and filtObserverName != self._observerName:
            return

        # If this message is an error response, then don't process it.
        if payloadEnvelope.result not in (None, True):
            logger.error("Vortex responded with error : %s" % payloadEnvelope.result)
            logger.error(str(payloadEnvelope.filt))
            return

        tupleSelector: TupleSelector = payloadEnvelope.filt["tupleSelector"]

        if not self._hasTupleSelector(tupleSelector):
            return

        cache, requiredUpdate = self._updateCache(payloadEnvelope)
        if not requiredUpdate:
            return

        payload = yield Payload().fromEncodedPayloadDefer(cache.encodedPayload)
        cache.subject.on_next(payload.tuples)

    def _tellServerWeWantData(self, tupleSelectors: List[TupleSelector]):
        for tupleSelector in tupleSelectors:
            self._sendRequestToServer(
                PayloadEnvelope({"subscribe": True, "tupleSelector": tupleSelector})
            )

    def _sendRequestToServer(self, payload):
        payload.filt.update(self._sendFilt)
        d = VortexFactory.sendVortexMsg(vortexMsgs=payload.toVortexMsg(),
                                        destVortexName=self._destVortexName)
        d.addErrback(vortexLogFailure, logger, consumeError=True)

    def _sendUnsubscribeToServer(self, tupleSelector: TupleSelector):
        payload = PayloadEnvelope()
        payload.filt["tupleSelector"] = tupleSelector
        payload.filt["unsubscribe"] = True
        self._sendRequestToServer(payload)



"""
 * Created by Synerty Pty Ltd
 *
 * This software is open source, the MIT license applies.
 *
 * Website : http://www.synerty.com
 * Support : support@synerty.com
"""

import datetime
import json

from twisted.trial import unittest

from vortex.Tuple import TUPLE_TYPES_BY_NAME, TupleHash
from vortex.test.TestTuple import TestSubTuple, TestTuple


def makeTupleA():
    tuple_ = TestTuple()
    tuple_.aBoolTrue = True
    tuple_.aBoolFalse = False
    tuple_.aDate = datetime.datetime(2010, 4, 7, 2, 33, 19, 666)
    tuple_.aFloat = 2.56
    tuple_.aInt = 1231231
    tuple_.aString = 'test string from 345345345@$#%#$%#$%#$%#'
    tuple_.aSet = ['x', 2, False] # Sets are restored as Lists
    tuple_.aList = ['y', 3, True, 3.3]
    tuple_.aDict = {'a': 'test str', # Key strings are loaded as unicode
                    1: 1.234}

    subTuple = TestSubTuple()
    subTuple.subInt = 2048
    tuple_.aSubTuple = subTuple
    #
    tuple_.aListOfSubTuples = []
    for x in range(2):
        subTuple = TestSubTuple()
        subTuple.subInt = x
        tuple_.aListOfSubTuples.append(subTuple)

    return tuple_


class TuplePyTest(unittest.TestCase):
    PERFORMANCE_ITERATIONS = 10000

    def setUp(self):
        pass

    def tearDown(self):
        pass


    def testTupleToFromJson(self):
        # Create tuple
        origTuple = makeTupleA()

        # To Json
        jsonDictStr = json.dumps(origTuple.toJsonDict())
        deserialisedTuple = TestTuple().fromJsonDict(json.loads(jsonDictStr))
        # print xmlStr

        self.assertEqual(TupleHash(origTuple), TupleHash(origTuple))
        self.assertEqual(TupleHash(deserialisedTuple), TupleHash(deserialisedTuple))

        # print TupleHash(origTuple)._key()
        # print TupleHash(deserialisedTuple)._key()

        self.assertEqual(TupleHash(origTuple), TupleHash(deserialisedTuple),
                         'Tuple serialise -> deserialise error')

    def testTupleToFromJsonPerformanceMulti(self):
        for _ in range(self.PERFORMANCE_ITERATIONS):
            origTuple = makeTupleA()
            jsonDictStr = json.dumps(origTuple.toJsonDict())
            TestTuple().fromJsonDict(json.loads(jsonDictStr))

    def testTupleToFromJsonPerformanceSingle(self):
        origTuple = makeTupleA()
        origTuple.aListOfSubTuples = [makeTupleA()
                                      for _ in range(self.PERFORMANCE_ITERATIONS)]
        jsonDictStr = json.dumps(origTuple.toJsonDict())
        TestTuple().fromJsonDict(json.loads(jsonDictStr))



    def testTupleHash(self):
        hashA1 = TupleHash(makeTupleA())
        hashA2 = TupleHash(makeTupleA())
        print("hashA1 = %s" % hashA1.__hash__())
        print("hashA2 = %s" % hashA2.__hash__())

        hashB1 = TupleHash(makeTupleA())
        hashB1.tupl.aString = 'Not the same as the others'
        print("hashB1 = %s" % hashB1.__hash__())

        hashA1.__hash__()
        self.assertTrue(hashA1.__hash__() == hashA1.__hash__(),
                        'TupleHash A1.__hash__() == A1.__hash__()')

        self.assertTrue(hashA1 == hashA2, 'TupleHash __eq__ fail, A1 == A2')
        self.assertFalse(hashA1 != hashA2, 'TupleHash __ne__ fail, A1 != A2')
        self.assertFalse(hashA1 == hashB1, 'TupleHash __eq__ fail, A1 == B1')
        self.assertTrue(hashA1 != hashB1, 'TupleHash __ne__ fail, A1 != B1')

        self.assertTrue(hashA1 in {hashA2: None}, 'TupleHash in fail, A1 in {A2:None}')
        self.assertFalse(hashA1 not in {hashA2: None},
                         'TupleHash in fail, A1 not in {A2:None}')
        self.assertTrue(hashA1 not in {hashB1: None},
                        'TupleHash in fail, A1 not in {B1:None}')
        self.assertFalse(hashA1 in {hashB1: None}, 'TupleHash in fail, A1 in {B1:None}')
